# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
import six
from six.moves import cPickle as pickle
from utool import util_path
from utool import util_inject
from os.path import splitext, basename, exists
try:
    import lockfile
    HAVE_LOCKFILE = True
except ImportError:
    HAVE_LOCKFILE = False
try:
    import numpy as np
    HAS_NUMPY = True
except ImportError:
    HAS_NUMPY = False

print, rrr, profile = util_inject.inject2(__name__)


__PRINT_IO__ = True
__PRINT_WRITES__ = __PRINT_IO__
__PRINT_READS__  =  __PRINT_IO__
__FORCE_PRINT_READS__ = False
__FORCE_PRINT_WRITES__ = False
__READ_TAIL_N__ = 3
#__FORCE_PRINT_READS__ = True
#__FORCE_PRINT_WRITES__ = True


def load_data(fpath, **kwargs):
    """ More generic interface to load data """
    ext = splitext(fpath)[1]
    if ext in ['.pickle', '.cPkl', '.pkl']:
        return load_cPkl(fpath, **kwargs)
    elif ext in ['.json']:
        return load_json(fpath, **kwargs)
    elif ext in ['.hdf5']:
        return load_hdf5(fpath, **kwargs)
    elif ext in ['.txt']:
        return load_text(fpath, **kwargs)
    elif HAS_NUMPY and ext in ['.npz', '.npy']:
        return load_numpy(fpath, **kwargs)
    else:
        assert False, 'unknown ext=%r for fpath=%r' % (ext, fpath)


def save_data(fpath, data, **kwargs):
    """ More generic interface to write data """
    ext = splitext(fpath)[1]
    if ext in ['.pickle', '.cPkl', '.pkl']:
        return save_cPkl(fpath, data, **kwargs)
    elif ext in ['.json']:
        return save_json(fpath, data, **kwargs)
    elif ext in ['.hdf5']:
        return save_hdf5(fpath, data, **kwargs)
    elif ext in ['.txt']:
        return save_text(fpath, **kwargs)
    elif HAS_NUMPY and ext in ['.npz', '.npy']:
        return save_numpy(fpath, data, **kwargs)
    else:
        assert False, 'unknown ext=%r for fpath=%r' % (ext, fpath)


def _rectify_verb_write(verbose):
    if __FORCE_PRINT_WRITES__:
        return True
    if verbose is None:
        verbose = __PRINT_WRITES__
    return verbose


def _rectify_verb_read(verbose):
    if __FORCE_PRINT_READS__:
        return True
    if verbose is None:
        verbose = __PRINT_READS__
    return verbose


def write_to(fpath, to_write, aslines=False, verbose=None,
             onlyifdiff=False, mode='w', n=None):
    """ Writes text to a file. Automatically encodes text as utf8.

    Args:
        fpath (str): file path
        to_write (str): text to write (must be unicode text)
        aslines (bool): if True to_write is assumed to be a list of lines
        verbose (bool): verbosity flag
        onlyifdiff (bool): only writes if needed!
                checks hash of to_write vs the hash of the contents of fpath
        mode (unicode): (default = u'w')
        n (int):  (default = 2)

    CommandLine:
        python -m utool.util_io --exec-write_to --show

    Ignore:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_io import *  # NOQA
        >>> import utool as ut
        >>> fpath = ut.unixjoin(ut.get_app_resource_dir('utool'), 'testwrite.txt')
        >>> ut.delete(fpath)
        >>> to_write = 'utf-8 symbols Δ, Й, ק, م, ๗, あ, 叶, 葉, and 말.'
        >>> aslines = False
        >>> verbose = True
        >>> onlyifdiff = False
        >>> mode = u'w'
        >>> n = 2
        >>> write_to(fpath, to_write, aslines, verbose, onlyifdiff, mode, n)
        >>> read_ = ut.read_from(fpath)
        >>> print('read_    = ' + read_)
        >>> print('to_write = ' + to_write)
        >>> assert read_ == to_write
    """
    if onlyifdiff:
        import utool as ut
        if ut.hashstr(read_from(fpath)) == ut.hashstr(to_write):
            print('[util_io] * no difference')
            return
    verbose = _rectify_verb_write(verbose)
    if verbose:
        # n = None if verbose > 1 else 2
        # print('[util_io] * Writing to text file: %r ' % util_path.tail(fpath, n=n))
        print('[util_io] * Writing to text file: {}'.format(fpath))

    backup = False and exists(fpath)
    if backup:
        util_path.copy(fpath, fpath + '.backup')

    if not isinstance(fpath, six.string_types):
        # Assuming a file object with a name attribute
        # Should just read from the file
        fpath = fpath.name

    with open(fpath, mode) as file_:
        if aslines:
            file_.writelines(to_write)
        else:
            # Ensure python2 writes in bytes
            if six.PY2:
                if isinstance(to_write, unicode):  # NOQA
                    to_write = to_write.encode('utf8')
            try:
                file_.write(to_write)
            except UnicodeEncodeError as ex:
                start = max(ex.args[2] - 10, 0)
                end = ex.args[3] + 10
                context = to_write[start:end]
                print(repr(context))
                print(context)
                from utool import util_dbg
                util_dbg.printex(ex, keys=[(type, 'to_write')])
                file_.close()
                if backup:
                    # restore
                    util_path.copy(fpath + '.backup', fpath)
                # import utool
                # utool.embed()
                raise


def read_from(fpath, verbose=None, aslines=False, strict=True, n=None, errors='replace'):
    r""" Reads text from a file. Automatically returns utf8.

    Args:
        fpath (str): file path
        aslines (bool): if True returns list of lines
        verbose (bool): verbosity flag

    Returns:
        str: text from fpath (this is unicode)

    Ignore:
        x = b'''/whaleshark_003_fors\xc3\xb8g.wmv" />\r\n'''
        ut.writeto('foo.txt', x)
        y = ut.readfrom('foo.txt')
        y.encode('utf8') == x
    """
    if n is None:
        n = __READ_TAIL_N__
    verbose = _rectify_verb_read(verbose)
    if verbose:
        print('[util_io] * Reading text file: %r ' % util_path.tail(fpath, n=n))
    try:
        if not util_path.checkpath(fpath, verbose=verbose, n=n):
            raise IOError('[io] * FILE DOES NOT EXIST!')
        #with open(fpath, 'r') as file_:
        with open(fpath, 'rb') as file_:
            if aslines:
                #text = file_.readlines()
                if six.PY2:
                    # python2 writes in bytes, so read as bytes then convert to
                    # utf8
                    text = [line.decode('utf8', errors=errors)
                            for line in file_.readlines()]
                else:
                    text = [line.decode('utf8', errors=errors)
                            for line in file_.readlines()]
                    #text = file_.readlines()
            else:
                # text = file_.read()
                if six.PY2:
                    text = file_.read().decode('utf8', errors=errors)
                else:
                    #text = file_.read()
                    text = file_.read().decode('utf8', errors=errors)
        return text
    except IOError as ex:
        from utool import util_dbg
        if verbose or strict:
            util_dbg.printex(ex, ' * Error reading fpath=%r' %
                             util_path.tail(fpath, n=n), '[io]')
        if strict:
            raise


def read_lines_from(fpath, num_lines=None, verbose=None, n=None):
    with open(fpath, 'r') as file_:
        line_list = []
        #for lineno, line in enumerate(file_.readline()):
        for lineno, line in enumerate(file_):
            line_list.append(line)
            if num_lines is not None and lineno > num_lines:
                break
    return line_list

# aliases
readfrom = read_from
writeto = write_to
save_text = write_to
load_text = read_from


def save_json(fpath, data, **kwargs):
    import utool as ut
    json_data = ut.to_json(data, **kwargs)
    ut.save_text(fpath, json_data)


def load_json(fpath):
    import utool as ut
    json_data = ut.load_text(fpath)
    data = ut.from_json(json_data)
    return data


def save_cPkl(fpath, data, verbose=None, n=None):
    """ Saves data to a pickled file with optional verbosity """
    verbose = _rectify_verb_write(verbose)
    if verbose:
        print('[util_io] * save_cPkl(%r, data)' % (util_path.tail(fpath, n=n),))
    with open(fpath, 'wb') as file_:
        # Use protocol 2 to support python2 and 3
        pickle.dump(data, file_, protocol=2)


def load_cPkl(fpath, verbose=None, n=None):
    r"""
    Loads a pickled file with optional verbosity.
    Aims for compatibility between python2 and python3.


    Ignore:
        >>> import utool as ut
        >>> def makedata_simple():
        >>>     data = np.empty((500, 2 ** 20), dtype=np.uint8) + 1
        >>>     return data
        >>> memtrack = ut.MemoryTracker()
        >>> # create a large amount of data
        >>> data = makedata_simple()
        >>> memtrack.report()
        >>> print(ut.get_object_size_str(data))
        >>> fpath = 'tmp.pkl'
        >>> ut.save_cPkl(fpath, data)
        >>> print(ut.get_file_nBytes_str('tmp.pkl'))
        >>> #del data
        >>> memtrack.collect()
        >>> memtrack.report()
        >>> data = ut.load_cPkl(fpath)
        >>> memtrack.report()

    Ignore:
        >>> def makedata_complex():
        >>>     rng = np.random.RandomState(42)
        >>>     item1 = np.empty((100, 2 ** 20), dtype=np.uint8) + 1
        >>>     item2 = [np.empty((10, 2 ** 10), dtype=np.uint8) + 1
        >>>              for a in range(1000)]
        >>>     item3 = {a: np.empty(int(rng.rand() * 10), dtype=np.int16) + 1
        >>>                  for a in range(100)}
        >>>     item4 = {np.int32(a): np.empty((int(rng.rand() * 10), 2), dtype=np.float64) + 1
        >>>                  for a in range(200)}
        >>>     data = {'item1': item1, 'item2': item2,
        >>>             'item3': item3, 'item4': item4}
        >>>     return data
        >>> memtrack = ut.MemoryTracker()
        >>> # create a large amount of data
        >>> data = makedata_complex()
        >>> memtrack.report()
        >>> print(ut.get_object_size_str(data))
        >>> fpath = 'tmp.pkl'
        >>> ut.save_cPkl(fpath, data)
        >>> print(ut.get_file_nBytes_str('tmp.pkl'))
        >>> #del data
        >>> memtrack.collect()
        >>> memtrack.report()
        >>> data2 = ut.load_cPkl(fpath)
        >>> memtrack.report()

    Ignore:
        >>> import utool as ut
        >>> memtrack = ut.MemoryTracker()
        >>> cacher = ut.Cacher('tmp', cache_dir='.', cfgstr='foo')
        >>> data3 = cacher.ensure(makedata_complex)
        >>> memtrack.report()
        >>> data4 = cacher.ensure(makedata_complex)
        >>> memtrack.report()
        >>> import utool as ut
        >>> memtrack = ut.MemoryTracker()
        >>> fpath = '/home/joncrall/Desktop/smkcache/inva_PZ_Master1VUUIDS((5616)vxihbjwtggyovrto)_vpgwpcafbjkkpjdf.cPkl'
        >>> print(ut.get_file_nBytes_str(fpath))
        >>> data = ut.load_cPkl(fpath)
        >>> memtrack.report()


        def makedata_complex():
            data = np.empty((1000, 2 ** 20), dtype=np.uint8)
            data[:] = 0
            return data
    """
    verbose = _rectify_verb_read(verbose)
    if verbose:
        print('[util_io] * load_cPkl(%r)' % (util_path.tail(fpath, n=n),))
    try:
        with open(fpath, 'rb') as file_:
            data = pickle.load(file_)
    except UnicodeDecodeError:
        if six.PY3:
            # try to open python2 pickle
            with open(fpath, 'rb') as file_:
                data = pickle.load(file_, encoding='latin1')
        else:
            raise
    except ValueError as ex:
        if six.PY2:
            if ex.message == 'unsupported pickle protocol: 4':
                raise ValueError(
                    'unsupported Python3 pickle protocol 4 '
                    'in Python2 for fpath=%r' % (fpath,))
            else:
                raise
        else:
            raise
    return data


def _python2_load_cpkl(fpath):
    """
    References:
        https://stackoverflow.com/questions/41720952/unpickle-sklearn-tree-descisiontreeregressor-in-python-2-from-python3
    """
    from lib2to3.fixes.fix_imports import MAPPING
    import sys
    import pickle

    # MAPPING maps Python 2 names to Python 3 names. We want this in reverse.
    REVERSE_MAPPING = {}
    for key, val in MAPPING.items():
        REVERSE_MAPPING[val] = key

    # We can override the Unpickler and loads
    class Python_3_Unpickler(pickle.Unpickler):
        """Class for pickling objects from Python 3"""
        def find_class(self, module, name):
            if module in REVERSE_MAPPING:
                module = REVERSE_MAPPING[module]
            __import__(module)
            mod = sys.modules[module]
            klass = getattr(mod, name)
            return klass

    def load(fpath):
        with open(fpath, 'rb') as file_:
            data = Python_3_Unpickler(file_).load()
        return data


def lock_and_load_cPkl(fpath, verbose=False):
    with lockfile.LockFile(fpath + '.lock'):
        return load_cPkl(fpath, verbose)


def lock_and_save_cPkl(fpath, data, verbose=False):
    with lockfile.LockFile(fpath + '.lock'):
        return save_cPkl(fpath, data, verbose)


def save_hdf5(fpath, data, verbose=None, compression='lzf'):
    r"""
    Restricted save of data using hdf5. Can only save ndarrays and dicts of
    ndarrays.

    Args:
        fpath (str):
        data (ndarray):
        compression (str):
            DEFLATE/GZIP - standard
            LZF  - fast
            SHUFFLE - compression ratio
            FLETCHER32 - error detection
            Scale-offset - integer / float scaling and truncation
            SZIP - fast and patented

    CommandLine:
        python -m utool.util_io --test-save_hdf5

    References:
        http://docs.h5py.org/en/latest/quick.html
        http://docs.h5py.org/en/latest/mpi.html

    Ignore:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_io import *  # NOQA
        >>> import numpy as np
        >>> import utool as ut
        >>> rng = np.random.RandomState(0)
        >>> data = (rng.rand(100000, 128) * 255).astype(np.uint8).copy()
        >>> verbose = True
        >>> fpath = ut.unixjoin(ut.ensure_app_resource_dir('utool'), 'myfile.hdf5')
        >>> compression = 'lzf'
        >>> ut.delete(fpath)
        >>> save_hdf5(fpath, data, verbose, compression)
        >>> data2 = load_hdf5(fpath, verbose)
        >>> assert data is not data2
        >>> assert np.all(data == data2)
        >>> assert ut.delete(fpath)

    Ignore:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_io import *  # NOQA
        >>> import numpy as np
        >>> import utool as ut
        >>> rng = np.random.RandomState(0)
        >>> data = {'name': 'foobar', 'x': [1, 2, 3], 'y': np.array([3, 2, 1])}
        >>> ut.exec_funckw(save_hdf5, globals())
        >>> fpath = ut.unixjoin(ut.ensure_app_resource_dir('utool'), 'myfile2.hdf5')
        >>> ut.delete(fpath)
        >>> save_hdf5(fpath, data, verbose, compression)
        >>> data2 = load_hdf5(fpath, verbose)
        >>> assert data is not data2
        >>> assert all([np.all(data[key] == data2[key]) for key in data.keys()])
        >>> assert ut.delete(fpath)

    Ignore:
        >>> # DISABLE_DOCTEST
        >>> # cPkl / numpy seems to be faster with this initial implementation
        >>> import utool as ut
        >>> data = (rng.rand(1000000, 128) * 255).astype(np.uint8).copy()
        >>> print(ut.get_object_size_str(data))
        >>> del data
        >>> setup = ut.codeblock(
        >>>     '''
                import numpy as np
                import utool as ut
                rng = np.random.RandomState(0)
                fpath = ut.unixjoin(ut.ensure_app_resource_dir('utool'), 'io_test_data')
                data = (rng.rand(1000000, 128) * 255).astype(np.uint8).copy()
                #print(ut.get_object_size_str(data))
                ''')
        >>> # Test load time
        >>> stmt_list1 = ut.codeblock(
        >>>     '''
                ut.save_hdf5(fpath + '.hdf5', data, verbose=False, compression='gzip')
                ut.save_hdf5(fpath + '.hdf5', data, verbose=False, compression='lzf')
                ut.save_cPkl(fpath + '.cPkl', data, verbose=False)
                ut.save_numpy(fpath + '.npy', data, verbose=False)
                ut.save_pytables(fpath + '.tables', data, verbose=False)
                ''').split('\n')
        >>> ut.util_dev.timeit_compare(stmt_list1, setup, int(10))
        >>> # Test save time
        >>> stmt_list2 = ut.codeblock(
        >>>     '''
                ut.load_hdf5(fpath + '.hdf5', verbose=False)
                ut.load_cPkl(fpath + '.cPkl', verbose=False)
                ut.load_numpy(fpath + '.npy', verbose=False)
                ut.load_pytables(fpath + '.tables', verbose=False)
                ''').split('\n')
        >>> ut.util_dev.timeit_compare(stmt_list2, setup, int(10))
        >>> print('finished timeing')
        +----------------
        | TIMEIT COMPARE
        +----------------
        | iterations = 10
        | Input:
        |     | num | stmt
        |     |   0 | u"ut.save_hdf5(fpath + '.hdf5', data, verbose=False, compression='gzip')"
        |     |   1 | u"ut.save_hdf5(fpath + '.hdf5', data, verbose=False, compression='lzf')"
        |     |   2 | u"ut.save_cPkl(fpath + '.cPkl', data, verbose=False)"
        |     |   3 | u"ut.save_numpy(fpath + '.npy', data, verbose=False)"
        |     |   4 | u"ut.save_pytables(fpath + '.tables', data, verbose=False)"
        ...
        | Output:
        |    * PASSED: each statement produced the same result
        |    | num | total time | per loop | stmt
        |    |   0 |    0.03 ks |   3.15 s | ut.save_hdf5(fpath + '.hdf5', data, verbose=False, compression='gzip')
        |    |   1 |    0.01 ks |   1.25 s | ut.save_hdf5(fpath + '.hdf5', data, verbose=False, compression='lzf')
        |    |   2 |     5.30 s |   0.53 s | ut.save_cPkl(fpath + '.cPkl', data, verbose=False)
        |    |   3 |     4.97 s |   0.50 s | ut.save_numpy(fpath + '.npy', data, verbose=False)
        |    |   4 |     9.23 s |   0.92 s | ut.save_pytables(fpath + '.tables', data, verbose=False)
        L_________________
        +----------------
        | TIMEIT COMPARE
        +----------------
        | iterations = 10
        | Input:
        |     | num | stmt
        |     |   0 | u"ut.load_hdf5(fpath + '.hdf5', verbose=False)"
        |     |   1 | u"ut.load_cPkl(fpath + '.cPkl', verbose=False)"
        |     |   2 | u"ut.load_numpy(fpath + '.npy', verbose=False)"
        |     |   3 | u"ut.load_pytables(fpath + '.tables', verbose=False)"
        ...
        | Output:
        |    * PASSED: each statement produced the same result
        |    | num | total time | per loop | stmt
        |    |   0 |     2.39 s |   0.24 s | ut.load_hdf5(fpath + '.hdf5', verbose=False)
        |    |   1 |     0.39 s |   0.04 s | ut.load_cPkl(fpath + '.cPkl', verbose=False)
        |    |   2 |     0.19 s |   0.02 s | ut.load_numpy(fpath + '.npy', verbose=False)
        |    |   3 |     0.33 s |   0.03 s | ut.load_pytables(fpath + '.tables', verbose=False)
        L_________________

    Ignore:
        %timeit save_hdf5(fpath, data, verbose=False, compression='gzip')
        %timeit save_hdf5(fpath, data, verbose=False, compression='lzf')
        %timeit save_cPkl(fpath + '.cPkl', data, verbose=False)
        %timeit save_pytables(fpath + '.tables', data, verbose=False)
        1 loops, best of 3: 258 ms per loop
        10 loops, best of 3: 111 ms per loop
        10 loops, best of 3: 53.1 ms per loop
        10 loops, best of 3: 96.5 ms per loop

        save_hdf5(fpath, data, verbose=False, compression='gzip')
        %timeit load_hdf5(fpath, verbose=False)
        save_hdf5(fpath, data, verbose=False, compression='lzf')
        %timeit load_hdf5(fpath, verbose=False)
        %timeit load_cPkl(fpath + '.cPkl', verbose=False)
        %timeit load_pytables(fpath + '.tables', verbose=False)
        100 loops, best of 3: 19.4 ms per loop
        100 loops, best of 3: 14.4 ms per loop
        100 loops, best of 3: 3.92 ms per loop
        100 loops, best of 3: 6.22 ms per loop

    Notes:
        pip install mpi4py
    """
    import h5py

    verbose = _rectify_verb_write(verbose)
    if verbose:
        print('[util_io] * save_hdf5(%r, data)' % (util_path.tail(fpath),))
    if verbose > 1:
        if isinstance(data, dict):
            print('[util_io] ... shapes=%r' % ([val.shape for val in data.values()],))
        else:
            print('[util_io] ... shape=%r' % (data.shape,))

    chunks = True  # True enables auto-chunking
    fname = basename(fpath)

    # check for parallel hdf5
    #have_mpi = h5py.h5.get_config().mpi
    #if have_mpi:
    #    import mpi4py
    #    h5kw = dict(driver='mpio', comm=mpi4py.MPI.COMM_WORLD)
    #    # cant use compression with mpi
    #    #ValueError: Unable to create dataset (Parallel i/o does not support filters yet)
    #else:
    h5kw = {}

    if isinstance(data, dict):
        array_data = {key: val for key, val in data.items()
                      if isinstance(val, (list, np.ndarray))}
        attr_data = {key: val for key, val in data.items() if key not in array_data}

        #assert all([
        #    isinstance(vals, np.ndarray)
        #    for vals in six.itervalues(data)
        #]), ('can only save dicts as ndarrays')
        # file_ = h5py.File(fpath, 'w', **h5kw)
        with h5py.File(fpath, mode='w', **h5kw) as file_:
            grp = file_.create_group(fname)
            for key, val in six.iteritems(array_data):
                val = np.asarray(val)
                dset = grp.create_dataset(
                    key, val.shape,  val.dtype, chunks=chunks,
                    compression=compression)
                dset[...] = val
            for key, val in six.iteritems(attr_data):
                grp.attrs[key] = val
    else:
        assert isinstance(data, np.ndarray)
        shape = data.shape
        dtype = data.dtype
        #if verbose or (verbose is None and __PRINT_WRITES__):
        #    print('[util_io] * save_hdf5(%r, data)' % (util_path.tail(fpath),))
        # file_ = h5py.File(fpath, 'w', **h5kw)
        with h5py.File(fpath, mode='w', **h5kw) as file_:
            #file_.create_dataset(
            #    fname, shape,  dtype, chunks=chunks, compression=compression,
            #    data=data)
            dset = file_.create_dataset(
                fname, shape,  dtype, chunks=chunks, compression=compression)
            dset[...] = data


def load_hdf5(fpath, verbose=None):
    import h5py
    fname = basename(fpath)
    #file_ = h5py.File(fpath, 'r')
    #file_.values()
    #file_.keys()
    verbose = _rectify_verb_read(verbose)
    if verbose:
        print('[util_io] * load_hdf5(%r)' % (util_path.tail(fpath),))
    with h5py.File(fpath, 'r') as file_:
        value = file_[fname]
        if isinstance(value, h5py.Group):
            grp = value
            data = {}
            for key, dset in six.iteritems(grp):
                shape = dset.shape
                dtype = dset.dtype
                subdata = np.empty(shape, dtype=dtype)
                dset.read_direct(subdata)
                data[key] = subdata
            for key, val in six.iteritems(grp.attrs):
                data[key] = val
        elif isinstance(value, h5py.Dataset):
            dset = value
            shape = dset.shape
            dtype = dset.dtype
            data = np.empty(shape, dtype=dtype)
            dset.read_direct(data)
        else:
            assert False
    return data


def save_pytables(fpath, data, verbose=False):
    """
    sudo pip install numexpr
    sudo pip install tables

    References:
        https://pytables.github.io/cookbook/py2exe_howto.html
        https://gist.github.com/andrewgiessel/7515520
        http://stackoverflow.com/questions/8843062/python-how-to-store-a-numpy-multidimensional-array-in-pytables
        http://pytables.github.io/usersguide/tutorials.html#creating-new-array-objects

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_io import *  # NOQA
        >>> import numpy as np
        >>> import utool as ut
        >>> # build test data
        >>> verbose = True
        >>> fpath = 'myfile.pytables.hdf5'
        >>> np.random.seed(0)
        >>> compression = 'gzip'
        >>> data = (np.random.rand(100000, 128) * 255).astype(np.uint8).copy()
        >>> # execute function
        >>> ut.delete(fpath)
        >>> save_pytables(fpath, data, verbose)
        >>> data2 = load_pytables(fpath, verbose)
        >>> assert data is not data2
        >>> assert np.all(data == data2)
        >>> assert ut.delete(fpath)
    """
    import tables
    #from os.path import basename
    #fname = basename(fpath)
    #shape = data.shape
    #dtype = data.dtype
    #file_ = tables.open_file(fpath)
    verbose = _rectify_verb_write(verbose)
    if verbose:
        print('[util_io] * save_pytables(%r, data)' % (util_path.tail(fpath),))
    with tables.open_file(fpath, 'w') as file_:
        atom = tables.Atom.from_dtype(data.dtype)
        filters = tables.Filters(complib='blosc', complevel=5)
        dset = file_.createCArray(file_.root, 'data', atom, data.shape, filters=filters)
        # save w/o compressive filter
        #dset = file_.createCArray(file_.root, 'all_data', atom, all_data.shape)
        dset[:] = data


def load_pytables(fpath, verbose=False):
    import tables
    #from os.path import basename
    #fname = basename(fpath)
    #file_ = tables.open_file(fpath)
    verbose = _rectify_verb_read(verbose)
    if verbose:
        print('[util_io] * load_pytables(%r, data)' % (util_path.tail(fpath),))
    with tables.open_file(fpath, 'r') as file_:
        data = file_.root.data.read()
    return data


def load_numpy(fpath, mmap_mode=None, verbose=None):
    verbose = _rectify_verb_read(verbose)
    if verbose:
        print('[util_io] * load_numpy(%r)' % util_path.tail(fpath))
    return np.load(fpath, mmap_mode=mmap_mode)


def save_numpy(fpath, data, verbose=None, **kwargs):
    verbose = _rectify_verb_write(verbose)
    if verbose:
        print('[util_io] * save_numpy(%r, data)' % util_path.tail(fpath))
    return np.save(fpath, data)


#def save_capnp(fpath, data, verbose=False):
#    r"""
#    Refernces:
#        http://jparyani.github.io/pycapnp/quickstart.html#dictionaries
#    """
#    import capnp
#    if verbose or __PRINT_WRITES__:
#        print('[util_io] * save_capnp(%r, data)' % (util_path.tail(fpath),))


def try_decode(x):
    # All python encoding formats
    codec_list = [
        'ascii', 'big5', 'big5hkscs', 'cp037', 'cp424', 'cp437', 'cp500',
        'cp720', 'cp737', 'cp775', 'cp850', 'cp852', 'cp855', 'cp856', 'cp857',
        'cp858', 'cp860', 'cp861', 'cp862', 'cp863', 'cp864', 'cp865', 'cp866',
        'cp869', 'cp874', 'cp875', 'cp932', 'cp949', 'cp950', 'cp1006',
        'cp1026', 'cp1140', 'cp1250', 'cp1251', 'cp1252', 'cp1253', 'cp1254',
        'cp1255', 'cp1256', 'cp1257', 'cp1258', 'euc_jp', 'euc_jis_2004',
        'euc_jisx0213', 'euc_kr', 'gb2312', 'gbk', 'gb18030', 'hz',
        'iso2022_jp', 'iso2022_jp_1', 'iso2022_jp_2', 'iso2022_jp_2004',
        'iso2022_jp_3', 'iso2022_jp_ext', 'iso2022_kr', 'latin_1', 'iso8859_2',
        'iso8859_3', 'iso8859_4', 'iso8859_5', 'iso8859_6', 'iso8859_7',
        'iso8859_8', 'iso8859_9', 'iso8859_10', 'iso8859_13', 'iso8859_14',
        'iso8859_15', 'iso8859_16', 'johab', 'koi8_r', 'koi8_u', 'mac_cyrillic',
        'mac_greek', 'mac_iceland', 'mac_latin2', 'mac_roman', 'mac_turkish',
        'ptcp154', 'shift_jis', 'shift_jis_2004', 'shift_jisx0213', 'utf_32',
        'utf_32_be', 'utf_32_le', 'utf_16', 'utf_16_be', 'utf_16_le', 'utf_7',
        'utf_8', 'utf_8_sig', ]
    for codec in codec_list:
        try:
            print(('%20s: ' % (codec,)) + x.encode(codec))
        except Exception:
            print(('%20s: ' % (codec,)) + 'FAILED')

    for codec in codec_list:
        try:
            print(('%20s: ' % (codec,)) + x.decode(codec))
        except Exception:
            print(('%20s: ' % (codec,)) + 'FAILED')


if __name__ == '__main__':
    """
    CommandLine:
        python -m utool.util_io
        python -m utool.util_io --allexamples
        python -m utool.util_io --allexamples --noface --nosrc
    """
    import multiprocessing
    multiprocessing.freeze_support()  # for win32
    import utool as ut  # NOQA
    ut.doctest_funcs()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
import types
import six
import sys

DEBUG        = '--debug' in sys.argv
DEBUG2       = '--debug2' in sys.argv
NO_ASSERTS   = '--no-assert' in sys.argv
QUIET        = '--quiet' in sys.argv
SILENT       = '--silent' in sys.argv
SAFE         = '--safe' in sys.argv
STRICT       = '--strict' not in sys.argv
REPORT       = '--report' not in sys.argv
TRACE        = '--trace' in sys.argv
SUPER_STRICT = '--super-strict' in sys.argv or '--superstrict' in sys.argv
VERBOSE      = '--verbose' in sys.argv or '--verb' in sys.argv
VERYVERBOSE  = '--very-verbose' in sys.argv or '--veryverbose' in sys.argv
NO_INDENT    = '--noindent' in sys.argv or '--no-indent' in sys.argv or SILENT
PRINT_ALL_CALLERS  = '--print-all-callers' in sys.argv
USE_ASSERT         = not NO_ASSERTS
NOT_QUIET          = not QUIET
PRINT_INJECT_ORDER = (VERYVERBOSE or '--print-inject-order' in sys.argv or
                      '--verbinject' in sys.argv or
                      '--print-imports' in sys.argv or
                      '--verb-inject' in sys.argv or
                      '--verbimport' in sys.argv or
                      '--verbimp' in sys.argv or
                      '--verb-import' in sys.argv or
                      '--verb-import' in sys.argv or
                      '--verbose-import' in sys.argv)
LOGGING_VERBOSE    = VERYVERBOSE or '--verb-logging' in sys.argv

if PRINT_INJECT_ORDER:
    # HACK
    from utool._internal import meta_util_dbg
    from six.moves import builtins
    N = 0
    callername = meta_util_dbg.get_caller_name(N=2 + N, strict=False)
    lineno = meta_util_dbg.get_caller_lineno(N=2 + N, strict=False)
    fmtdict = dict(N=N, lineno=lineno, callername=callername, modname=__name__)
    msg = '[util_inject] N={N} {modname} is imported by {callername} at lineno={lineno}'.format(**fmtdict)
    builtins.print(msg)


def _try_cast(val, type_):
    if type_ in [types.BooleanType] and isinstance(val, six.string_types):
        if val.lower() == 'true':
            newval = True
        elif val.lower() == 'false':
            newval = True
    else:
        newval = type_(val)
    return newval


def get_argflag(flag):
    return flag in sys.argv


# FIXME: on a rrrr this get_argval is taken
# instead of the correct one from util_arg
def get_argval(argstr, type_=None, default=None):
    r"""
    Returns a value of an argument specified on the command line after some flag

    CommandLine:
        python -c "import utool; print([(type(x), x) for x in [utool.get_argval('--quest')]])" --quest="holy grail"
        python -c "import utool; print([(type(x), x) for x in [utool.get_argval('--quest')]])" --quest="42"
        python -c "import utool; print([(type(x), x) for x in [utool.get_argval('--quest')]])" --quest=42
        python -c "import utool; print([(type(x), x) for x in [utool.get_argval('--quest')]])" --quest 42
        python -c "import utool; print([(type(x), x) for x in [utool.get_argval('--quest', float)]])" --quest 42

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_arg import *  # NOQA
        >>> import sys
        >>> sys.argv.extend(['--spam', 'eggs', '--quest=holy grail', '--ans=42'])
        >>> get_argval('--spam', type_=str, default=None)
        eggs
        >>> get_argval('--quest', type_=str, default=None)
        holy grail
        >>> get_argval('--ans', type_=int, default=None)
        42
    """
    arg_after = default
    if type_ is bool:
        arg_after = False if default is None else default
    try:
        # New for loop way (accounts for =)
        for argx, item in enumerate(sys.argv):
            if item == argstr:
                if argx < len(sys.argv):
                    if type_ is bool:
                        arg_after = True
                    else:
                        arg_after = _try_cast(sys.argv[argx + 1], type_)
            if item.startswith(argstr + '='):
                val_after = ''.join(item.split('=')[1:])
                arg_after = _try_cast(val_after, type_)
    except Exception:
        pass
    return arg_after



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals


def utool_main():
    ignore_prefix = []
    ignore_suffix = []
    import utool as ut
    # allows for --tf
    ut.main_function_tester('utool', ignore_prefix, ignore_suffix)

if __name__ == '__main__':
    """
    python -m utool --tf infer_function_info:0
    """
    print('Checking utool main')
    utool_main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
autogenerates sphinx documentation
"""
from __future__ import absolute_import, division, print_function, unicode_literals
import utool as ut


if __name__ == '__main__':
    """
    autogen_sphinx_apidoc
    OR
    python -m utool.util_setup --exec-autogen_sphinx_apidoc
    """
    ut.autogen_sphinx_apidoc()



# -*- coding: utf-8 -*-
"""
If logging is on, utool will overwrite the print function with a logging function

This is a special module which will not get injected into (should it be internal?)

References:
    # maybe we can do something like this Queue to try fixing error when
    # when using injected print statments with Qt signals and slots
    http://stackoverflow.com/questions/21071448/redirecting-stdout-and-stderr-to-a-pyqt4-qtextedit-from-a-secondary-thread
"""
from __future__ import absolute_import, division, print_function, unicode_literals
import six
from six.moves import builtins, map, zip, range  # NOQA
from os.path import exists, join, realpath
import logging
import logging.config
import multiprocessing
import os
import sys
from utool._internal import meta_util_arg, meta_util_six

VERBOSE            = meta_util_arg.VERBOSE
VERYVERBOSE        = meta_util_arg.VERYVERBOSE
PRINT_ALL_CALLERS  = meta_util_arg.PRINT_ALL_CALLERS
LOGGING_VERBOSE    = meta_util_arg.LOGGING_VERBOSE  # --verb-logging
PRINT_INJECT_ORDER = meta_util_arg.PRINT_INJECT_ORDER


def __inside_doctest(original_stdout=sys.stdout):
    return original_stdout != sys.stdout


__IN_MAIN_PROCESS__ = multiprocessing.current_process().name == 'MainProcess'

__UTOOL_ROOT_LOGGER__ = None
__CURRENT_LOG_FPATH__ = None

# Remeber original python values
# __PYTHON_STDOUT__ = sys.stdout
# __PYTHON_PRINT__  = builtins.print
# __PYTHON_WRITE__  = __PYTHON_STDOUT__.write
# __PYTHON_FLUSH__  = __PYTHON_STDOUT__.flush

# Initialize utool values
__UTOOL_STDOUT__    = None
__UTOOL_PRINT__     = None

# TODO: Allow write and flush to have a logging equivalent
__UTOOL_WRITE__     = None
__UTOOL_FLUSH__     = None
__UTOOL_WRITE_BUFFER__ = []


def _utool_stdout():
    if __UTOOL_STDOUT__ is not None:
        return __UTOOL_STDOUT__
    else:
        return sys.stdout


def _utool_write():
    if __UTOOL_WRITE__ is not None:
        return __UTOOL_WRITE__
    else:
        return sys.stdout.write


def _utool_flush():
    if __UTOOL_FLUSH__ is not None:
        return __UTOOL_FLUSH__
    else:
        return sys.stdout.flush


def _utool_print():
    if __UTOOL_PRINT__ is not None:
        return __UTOOL_PRINT__
    else:
        return builtins.print


__STR__ = six.text_type

logdir_cacheid = 'log_dpath'


def testlogprog():
    r"""
    Test to ensure that all progress lines are outputed to the file logger
    while only a few progress lines are outputed to stdout.
    (if backspace is specified)

    CommandLine:
        python -m utool.util_logging testlogprog --show --verb-logging
        python -m utool.util_logging testlogprog --show

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_logging import *  # NOQA
        >>> import utool as ut
        >>> result = testlogprog()
        >>> print(result)
    """
    import utool as ut
    print('Starting test log function')

    def test_body(count, logmode, backspace):
        ut.colorprint('\n---- count = %r -----' % (count,), 'yellow')
        ut.colorprint('backspace = %r' % (backspace,), 'yellow')
        ut.colorprint('logmode = %r' % (logmode,), 'yellow')
        if logmode:
            ut.delete('test.log')
            ut.start_logging('test.log')
        print('Start main loop')
        import time
        for count in ut.ProgressIter(range(20), freq=3, backspace=backspace):
            time.sleep(.01)
        print('Done with main loop work')
        print('Exiting main body')
        if logmode:
            ut.stop_logging()
            #print('-----DONE LOGGING----')
            testlog_text = ut.readfrom('test.log')
            print(ut.indent(testlog_text.replace('\r', '\n'), '        '))

    def test_body2(count, logmode, backspace):
        ut.colorprint('\n---- count = %r -----' % (count,), 'yellow')
        ut.colorprint('backspace = %r' % (backspace,), 'yellow')
        ut.colorprint('logmode = %r' % (logmode,), 'yellow')
        if logmode:
            ut.delete('test.log')
            ut.start_logging('test.log')
        print('Start main loop')
        import time
        for count in ut.ProgressIter(range(2), freq=1, backspace=backspace):
            for count in ut.ProgressIter(range(50), freq=1, backspace=backspace):
                time.sleep(.01)
        print('Done with main loop work')
        print('Exiting main body')
        if logmode:
            ut.stop_logging()
            #print('-----DONE LOGGING----')
            #testlog_text = ut.readfrom('test.log')
            #print(ut.indent(testlog_text.replace('\r', '\n'), '        '))

    #test_body(0, False, True)
    #test_body(1, False, False)
    #test_body(2, True, True)
    #test_body(3, True, False)

    test_body2(4, True, True)
    test_body2(5, False, True)


def ensure_logging():
    flag = is_logging()
    if not flag:
        start_logging()
    return flag


def is_logging():
    global __UTOOL_ROOT_LOGGER__
    flag = __UTOOL_ROOT_LOGGER__ is not None
    return flag


def debug_logging_iostreams():
    print(' --- <DEBUG IOSTREAMS> --')
    print('__STR__ = %r' % (__STR__,))
    print('__IN_MAIN_PROCESS__ = %r' % (__IN_MAIN_PROCESS__,))
    print('__UTOOL_ROOT_LOGGER__ = %r' % (__UTOOL_ROOT_LOGGER__,))
    print('__CURRENT_LOG_FPATH__ = %r' % (__CURRENT_LOG_FPATH__,))
    # print('__PYTHON_STDOUT__ = %r' % (__PYTHON_STDOUT__,))
    # print('__PYTHON_PRINT__ = %r' % (__PYTHON_PRINT__,))
    # print('__PYTHON_WRITE__ = %r' % (__PYTHON_WRITE__,))
    # print('__PYTHON_FLUSH__ = %r' % (__PYTHON_FLUSH__,))
    print('__UTOOL_STDOUT__ = %r' % (__UTOOL_STDOUT__,))
    print('__UTOOL_PRINT__ = %r' % (__UTOOL_PRINT__,))
    print('__UTOOL_FLUSH__ = %r' % (__UTOOL_FLUSH__,))
    print('__UTOOL_WRITE__ = %r' % (__UTOOL_WRITE__,))
    print(' --- </DEBUG IOSTREAMS> --')


def get_logging_dir(appname='default'):
    """
    The default log dir is in the system resource directory
    But the utool global cache allows for the user to override
    where the logs for a specific app should be stored.

    Returns:
        log_dir_realpath (str): real path to logging directory
    """
    from utool._internal import meta_util_cache
    from utool._internal import meta_util_cplat
    from utool import util_cache
    if appname is None or  appname == 'default':
        appname = util_cache.get_default_appname()
    resource_dpath = meta_util_cplat.get_resource_dir()
    default = join(resource_dpath, appname, 'logs')
    # Check global cache for a custom logging dir otherwise
    # use the default.
    log_dir = meta_util_cache.global_cache_read(logdir_cacheid,
                                                appname=appname,
                                                default=default)
    log_dir_realpath = realpath(log_dir)
    return log_dir_realpath


def get_shelves_dir(appname='default'):
    """
    The default shelf dir is in the system resource directory
    But the utool global cache allows for the user to override
    where the shelf for a specific app should be stored.

    Returns:
        log_dir_realpath (str): real path to shelves directory
    """
    from utool._internal import meta_util_cache
    from utool._internal import meta_util_cplat
    from utool import util_cache
    if appname is None or  appname == 'default':
        appname = util_cache.get_default_appname()
    resource_dpath = meta_util_cplat.get_resource_dir()
    default = join(resource_dpath, appname, 'shelves')
    # Check global cache for a custom logging dir otherwise
    # use the default.
    log_dir = meta_util_cache.global_cache_read(logdir_cacheid,
                                                appname=appname,
                                                default=default)
    log_dir_realpath = realpath(log_dir)
    return log_dir_realpath


def get_current_log_fpath():
    global __CURRENT_LOG_FPATH__
    return __CURRENT_LOG_FPATH__


def get_current_log_text():
    fpath = get_current_log_fpath()
    if fpath is None:
        text = None
    else:
        with open(fpath, 'r') as file_:
            text = file_.read()
    return text


def get_log_fpath(num='next', appname=None, log_dir=None):
    """
    Returns:
        log_fpath (str): path to log file
    """
    if log_dir is None:
        log_dir = get_logging_dir(appname=appname)
    if not exists(log_dir):
        os.makedirs(log_dir)
    if appname is not None:
        log_fname = appname + '_logs_%04d.out'
    else:
        log_fname = 'utool_logs_%04d.out'
    if isinstance(num, six.string_types):
        if num == 'next':
            count = 0
            log_fpath = join(log_dir, log_fname % count)
            while exists(log_fpath):
                log_fpath = join(log_dir, log_fname % count)
                count += 1
    else:
        log_fpath = join(log_dir, log_fname % num)
    return log_fpath


def get_utool_logger():
    return __UTOOL_ROOT_LOGGER__


def add_logging_handler(handler, format_='file'):
    """
    mostly for util_logging internals
    """
    global __UTOOL_ROOT_LOGGER__
    if __UTOOL_ROOT_LOGGER__ is None:
        builtins.print('[WARNING] logger not started, cannot add handler')
        return
    # create formatter and add it to the handlers
    #logformat = '%Y-%m-%d %H:%M:%S'
    #logformat = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
    timeformat = '%H:%M:%S'
    if format_ == 'file':
        logformat = '[%(asctime)s]%(message)s'
    elif format_ == 'stdout':
        logformat = '%(message)s'
    else:
        raise AssertionError('unknown logging format_: %r' % format_)
    # Create formatter for handlers
    formatter = logging.Formatter(logformat, timeformat)
    handler.setLevel(logging.DEBUG)
    handler.setFormatter(formatter)
    __UTOOL_ROOT_LOGGER__.addHandler(handler)


class CustomStreamHandler(logging.Handler):
    """
    Modified from logging.py
    """

    def __init__(self, stream=None):
        """
        Initialize the handler.

        If stream is not specified, sys.stderr is used.
        """
        self.terminator = "\n"
        logging.Handler.__init__(self)
        if stream is None:
            stream = sys.stderr
        self.stream = stream

    def flush(self):
        """
        Flushes the stream.
        """
        self.acquire()
        try:
            if self.stream and hasattr(self.stream, "flush"):
                self.stream.flush()
        finally:
            self.release()

    def emit(self, record):
        """
        Emit a record.

        If a formatter is specified, it is used to format the record.
        The record is then written to the stream with a trailing newline.  If
        exception information is present, it is formatted using
        traceback.print_exception and appended to the stream.  If the stream
        has an 'encoding' attribute, it is used to determine how to do the
        output to the stream.
        """
        try:
            msg = self.format(record)
            stream = self.stream
            fs = "%s%s"
            if six.PY3 or not logging._unicode:  # if no unicode support...
                stream.write(fs % (msg, self.terminator))
            else:
                try:
                    if (isinstance(msg, six.text_type) and getattr(stream, 'encoding', None)):
                        ufs = u'%s%s'
                        try:
                            stream.write(ufs % (msg, self.terminator))
                        except UnicodeEncodeError:
                            #Printing to terminals sometimes fails. For example,
                            #with an encoding of 'cp1251', the above write will
                            #work if written to a stream opened or wrapped by
                            #the codecs module, but fail when writing to a
                            #terminal even when the codepage is set to cp1251.
                            #An extra encoding step seems to be needed.
                            stream.write((ufs % (msg, self.terminator)).encode(stream.encoding))
                    else:
                        stream.write(fs % (msg, self.terminator))
                except UnicodeError:
                    stream.write(fs % (msg.encode("UTF-8"), self.terminator.encode("UTF-8")))
            #self.flush()
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception:
            self.handleError(record)


def start_logging(log_fpath=None, mode='a', appname='default', log_dir=None):
    r"""
    Overwrites utool print functions to use a logger

    CommandLine:
        python -m utool.util_logging --test-start_logging:0
        python -m utool.util_logging --test-start_logging:1

    Example0:
        >>> # DISABLE_DOCTEST
        >>> import sys
        >>> sys.argv.append('--verb-logging')
        >>> import utool as ut
        >>> ut.start_logging()
        >>> ut.util_logging._utool_print()('hello world')
        >>> ut.util_logging._utool_write()('writing1')
        >>> ut.util_logging._utool_write()('writing2\n')
        >>> ut.util_logging._utool_write()('writing3')
        >>> ut.util_logging._utool_flush()()
        >>> handler = ut.util_logging.__UTOOL_ROOT_LOGGER__.handlers[0]
        >>> current_log_fpath = handler.stream.name
        >>> current_log_text = ut.read_from(current_log_fpath)
        >>> print('current_log_text =\n%s' % (current_log_text,))
        >>> assert current_log_text.find('hello world') > 0, 'cant hello world'
        >>> assert current_log_text.find('writing1writing2') > 0, 'cant find writing1writing2'
        >>> assert current_log_text.find('writing3') > 0, 'cant find writing3'

    Example1:
        >>> # DISABLE_DOCTEST
        >>> # Ensure that progress is logged
        >>> import sys
        >>> sys.argv.append('--verb-logging')
        >>> import utool as ut
        >>> ut.start_logging()
        >>> [x for x in  ut.ProgressIter(range(0, 1000), freq=4)]
        >>> handler = ut.util_logging.__UTOOL_ROOT_LOGGER__.handlers[0]
        >>> current_log_fpath = handler.stream.name
        >>> current_log_text = ut.read_from(current_log_fpath)
        >>> assert current_log_text.find('rate') > 0, 'progress was not logged'
        >>> print(current_log_text)
    """
    global __UTOOL_ROOT_LOGGER__
    global __UTOOL_PRINT__
    global __UTOOL_WRITE__
    global __UTOOL_FLUSH__
    global __CURRENT_LOG_FPATH__
    if LOGGING_VERBOSE:
        print('[utool] start_logging()')
    # FIXME: The test for doctest may not work

    if __UTOOL_ROOT_LOGGER__ is None and __IN_MAIN_PROCESS__ and not __inside_doctest():
        if LOGGING_VERBOSE:
            print('[utool] start_logging()... rootcheck OK')
        #logging.config.dictConfig(LOGGING)
        if log_fpath is None:
            log_fpath = get_log_fpath(num='next', appname=appname, log_dir=log_dir)
        __CURRENT_LOG_FPATH__ = log_fpath
        # Print what is about to happen
        if VERBOSE or LOGGING_VERBOSE:
            startmsg = ('logging to log_fpath=%r' % log_fpath)
            _utool_print()(startmsg)
        # Create root logger
        __UTOOL_ROOT_LOGGER__ = logging.getLogger('root')
        __UTOOL_ROOT_LOGGER__.setLevel('DEBUG')
        # create file handler which logs even debug messages
        #fh = logging.handlers.WatchedFileHandler(log_fpath)
        logfile_handler = logging.FileHandler(log_fpath, mode=mode)
        #stdout_handler = logging.StreamHandler(__UTOOL_STDOUT__)
        stdout_handler = CustomStreamHandler(__UTOOL_STDOUT__)
        stdout_handler.terminator = ''
        # http://stackoverflow.com/questions/7168790/suppress-newline-in-python-logging-module
        #stdout_handler.terminator = ''
        add_logging_handler(logfile_handler, format_='file')
        add_logging_handler(stdout_handler, format_='stdout')
        __UTOOL_ROOT_LOGGER__.propagate = False
        __UTOOL_ROOT_LOGGER__.setLevel(logging.DEBUG)
        # Overwrite utool functions with the logging functions

        def utool_flush(*args):
            """ flushes whatever is in the current utool write buffer """
            # Flushes only the stdout handler
            stdout_handler.flush()
            #__UTOOL_ROOT_LOGGER__.flush()
            #global __UTOOL_WRITE_BUFFER__
            #if len(__UTOOL_WRITE_BUFFER__) > 0:
            #    msg = ''.join(__UTOOL_WRITE_BUFFER__)
            #    #sys.stdout.write('FLUSHING %r\n' % (len(__UTOOL_WRITE_BUFFER__)))
            #    __UTOOL_WRITE_BUFFER__ = []
            #    return __UTOOL_ROOT_LOGGER__.info(msg)
            #__PYTHON_FLUSH__()

        def utool_write(*args):
            """ writes to current utool logs and to sys.stdout.write """
            #global __UTOOL_WRITE_BUFFER__
            #sys.stdout.write('WRITEING\n')
            msg = ', '.join(map(six.text_type, args))
            #__UTOOL_WRITE_BUFFER__.append(msg)
            __UTOOL_ROOT_LOGGER__.info(msg)
            #if msg.endswith('\n'):
            #    # Flush on newline, and remove newline
            #    __UTOOL_WRITE_BUFFER__[-1] = __UTOOL_WRITE_BUFFER__[-1][:-1]
            #    utool_flush()
            #elif len(__UTOOL_WRITE_BUFFER__) > 32:
            #    # Flush if buffer is too large
            #    utool_flush()

        if not PRINT_ALL_CALLERS:
            def utool_print(*args):
                """ standard utool print function """
                #sys.stdout.write('PRINT\n')
                endline = '\n'
                try:
                    msg = ', '.join(map(six.text_type, args))
                    return  __UTOOL_ROOT_LOGGER__.info(msg + endline)
                except UnicodeDecodeError:
                    new_msg = ', '.join(map(meta_util_six.ensure_unicode, args))
                    #print(new_msg)
                    return  __UTOOL_ROOT_LOGGER__.info(new_msg + endline)
        else:
            def utool_print(*args):
                """ debugging utool print function """
                import utool as ut
                utool_flush()
                endline = '\n'
                __UTOOL_ROOT_LOGGER__.info('\n\n----------')
                __UTOOL_ROOT_LOGGER__.info(ut.get_caller_name(range(0, 20)))
                return  __UTOOL_ROOT_LOGGER__.info(', '.join(map(six.text_type, args)) + endline)

        def utool_printdbg(*args):
            """ DRPRICATE standard utool print debug function """
            return  __UTOOL_ROOT_LOGGER__.debug(', '.join(map(six.text_type, args)))
        # overwrite the utool printers
        __UTOOL_WRITE__    = utool_write
        __UTOOL_FLUSH__    = utool_flush
        __UTOOL_PRINT__    = utool_print
        # Test out our shiney new logger
        if VERBOSE or LOGGING_VERBOSE:
            __UTOOL_PRINT__('<__LOG_START__>')
            __UTOOL_PRINT__(startmsg)
    else:
        if LOGGING_VERBOSE:
            print('[utool] start_logging()... FAILED TO START')
            print('DEBUG INFO')
            print('__inside_doctest() = %r' % (__inside_doctest(),))
            print('__IN_MAIN_PROCESS__ = %r' % (__IN_MAIN_PROCESS__,))
            print('__UTOOL_ROOT_LOGGER__ = %r' % (__UTOOL_ROOT_LOGGER__,))


def stop_logging():
    """
    Restores utool print functions to python defaults
    """
    global __UTOOL_ROOT_LOGGER__
    global __UTOOL_PRINT__
    global __UTOOL_WRITE__
    global __UTOOL_FLUSH__
    if __UTOOL_ROOT_LOGGER__ is not None:
        # Flush remaining buffer
        if VERBOSE or LOGGING_VERBOSE:
            _utool_print()()('<__LOG_STOP__>')
        _utool_flush()()
        # Remove handlers
        for h in __UTOOL_ROOT_LOGGER__.handlers[:]:
            __UTOOL_ROOT_LOGGER__.removeHandler(h)
        # Reset objects
        __UTOOL_ROOT_LOGGER__ = None
        __UTOOL_PRINT__    = None
        __UTOOL_WRITE__    = None
        __UTOOL_FLUSH__    = None

# HAVE TO HACK THIS IN FOR UTOOL SPECIAL CASE ONLY
# OTHER MODULE CAN USE NOINJECT
if PRINT_INJECT_ORDER:
    from utool._internal import meta_util_dbg
    callername = meta_util_dbg.get_caller_name(N=1, strict=False)
    fmtdict = dict(callername=callername, modname='utool.util_logging')
    msg = '[util_inject] {modname} is imported by {callername}'.format(**fmtdict)
    builtins.print(msg)


if __name__ == '__main__':
    """
    CommandLine:
        python -m utool.util_logging
        python -m utool.util_logging --allexamples
        python -m utool.util_logging --allexamples --noface --nosrc
    """
    import multiprocessing
    multiprocessing.freeze_support()  # for win32
    import utool as ut  # NOQA
    ut.doctest_funcs()



# -*- coding: utf-8 -*-
"""
Aliases harder to remember or very common standard module functions
"""
from __future__ import absolute_import, division, print_function

from itertools import product, combinations
from collections import defaultdict, OrderedDict, namedtuple
from functools import partial

ddict = defaultdict
odict = OrderedDict
iprod = product
icomb = combinations
namedtuple = namedtuple
partial = partial



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function
import re
import os
import six
from collections import deque  # NOQA
from os.path import exists, dirname, join, expanduser, normpath
from utool import util_inject
print, rrr, profile = util_inject.inject2(__name__)


class PythonStatement(object):
    """ Thin wrapper around a string representing executable python code """
    def __init__(self, stmt):
        self.stmt = stmt
    def __repr__(self):
        return self.stmt
    def __str__(self):
        return self.stmt


def dump_autogen_code(fpath, autogen_text, codetype='python', fullprint=None,
                      show_diff=None, dowrite=None):
    """
    Helper that write a file if -w is given on command line, otherwise
    it just prints it out. It has the opption of comparing a diff to the file.
    """
    import utool as ut
    if dowrite is None:
        dowrite = ut.get_argflag(('-w', '--write'))
    if show_diff is None:
        show_diff = ut.get_argflag('--diff')
    num_context_lines = ut.get_argval('--diff', type_=int, default=None)
    show_diff = show_diff or num_context_lines is not None

    num_context_lines = ut.get_argval('--diff', type_=int, default=None)

    if fullprint is None:
        fullprint = True

    if fullprint is False:
        fullprint = ut.get_argflag('--print')

    print('[autogen] Autogenerated %s...\n+---\n' % (fpath,))
    if not dowrite:
        if fullprint:
            ut.print_code(autogen_text, lexer_name=codetype)
            print('\nL___')
        else:
            print('specify --print to write to stdout')
            pass
        print('specify -w to write, or --diff to compare')
        print('...would write to: %s' % fpath)
    if show_diff:
        if ut.checkpath(fpath, verbose=True):
            prev_text = ut.read_from(fpath)
            textdiff = ut.get_textdiff(prev_text, autogen_text,
                                       num_context_lines=num_context_lines)
            try:
                ut.print_difftext(textdiff)
            except UnicodeDecodeError:
                import unicodedata
                textdiff = unicodedata.normalize('NFKD', textdiff).encode('ascii', 'ignore')
                ut.print_difftext(textdiff)

        if dowrite:
            print('WARNING: Not writing. Remove --diff from command line')
    elif dowrite:
        ut.write_to(fpath, autogen_text)


def makeinit(mod_dpath, exclude_modnames=[], use_star=False):
    r"""
    Args:
        mod_dpath (str):
        exclude_modnames (list): (Defaults to [])
        use_star (bool): (Defaults to False)

    Returns:
        str: init_codeblock

    CommandLine:
        python -m utool.util_autogen makeinit --modname=ibeis.algo

    Example:
        >>> # DISABLE_DOCTEST
        >>> # SCRIPT
        >>> from utool.util_autogen import *  # NOQA
        >>> import utool as ut
        >>> modname = ut.get_argval('--modname', str, default=None)
        >>> mod_dpath = (os.getcwd() if modname is None else
        >>>              ut.get_modpath(modname, prefer_pkg=True))
        >>> mod_dpath = ut.unixpath(mod_dpath)
        >>> mod_fpath = join(mod_dpath, '__init__.py')
        >>> exclude_modnames = ut.get_argval(('--exclude', '-x'), list, default=[])
        >>> use_star = ut.get_argflag('--star')
        >>> init_codeblock = makeinit(mod_dpath, exclude_modnames, use_star)
        >>> ut.dump_autogen_code(mod_fpath, init_codeblock)
    """
    from utool._internal import util_importer
    import utool as ut
    module_name = ut.get_modname_from_modpath(mod_dpath)
    IMPORT_TUPLES = util_importer.make_import_tuples(mod_dpath, exclude_modnames=exclude_modnames)
    initstr = util_importer.make_initstr(module_name, IMPORT_TUPLES)
    regen_command = 'cd %s\n' % (mod_dpath)
    regen_command += '    makeinit.py'
    regen_command += ' --modname={modname}'.format(modname=module_name)
    if use_star:
        regen_command += ' --star'
    if len(exclude_modnames ) > 0:
        regen_command += ' -x ' + ' '.join(exclude_modnames)

    regen_block = (ut.codeblock('''
    """
    Regen Command:
        {regen_command}
    """
    ''').format(regen_command=regen_command))

    importstar_codeblock = ut.codeblock(
        '''
        """
        python -c "import {module_name}" --dump-{module_name}-init
        python -c "import {module_name}" --update-{module_name}-init
        """
        __DYNAMIC__ = True
        if __DYNAMIC__:
            # TODO: import all utool external prereqs. Then the imports will not import
            # anything that has already in a toplevel namespace
            # COMMENTED OUT FOR FROZEN __INIT__
            # Dynamically import listed util libraries and their members.
            from utool._internal import util_importer
            # FIXME: this might actually work with rrrr, but things arent being
            # reimported because they are already in the modules list
            import_execstr = util_importer.dynamic_import(__name__, IMPORT_TUPLES)
            exec(import_execstr)
            DOELSE = False
        else:
            # Do the nonexec import (can force it to happen no matter what if alwyas set
            # to True)
            DOELSE = True

        if DOELSE:
            # <AUTOGEN_INIT>
            pass
            # </AUTOGEN_INIT>
        '''.format(module_name=module_name)
    )

    ts_line = '# Autogenerated on {ts}'.format(ts=ut.get_timestamp('printable'))

    init_codeblock_list = ['# -*- coding: utf-8 -*-', ts_line]
    init_codeblock_list.append(initstr)
    init_codeblock_list.append('\nIMPORT_TUPLES = ' + ut.repr4(IMPORT_TUPLES))
    if use_star:
        init_codeblock_list.append(importstar_codeblock)
    init_codeblock_list.append(regen_block)

    init_codeblock = '\n'.join(init_codeblock_list)
    return init_codeblock


def write_modscript_alias(fpath, modname, args='', pyscript='python'):
    """
    convinience function because $@ is annoying to paste into the terminal
    """
    import utool as ut
    from os.path import splitext
    allargs_dict = {
        '.sh': ' $@',
        '.bat': ' %1', }
    _, script_ext = splitext(fpath)
    if script_ext not in ['.sh', '.bat']:
        script_ext = '.bat' if ut.WIN32 else 'sh'
    allargs = (args + allargs_dict[script_ext]).strip(' ')
    if not modname.endswith('.py'):
        fmtstr = '{pyscript} -m {modname} {allargs}'
    else:
        fmtstr = '{pyscript} {modname} {allargs}'

    cmdstr = fmtstr.format(pyscript=pyscript, modname=modname, allargs=allargs)
    ut.write_to(fpath, cmdstr)
    os.system('chmod +x ' + fpath)


def autofix_codeblock(codeblock, max_line_len=80,
                      aggressive=False,
                      very_aggressive=False,
                      experimental=False):
    r"""
    Uses autopep8 to format a block of code

    Example:
        >>> # DISABLE_DOCTEST
        >>> import utool as ut
        >>> codeblock = ut.codeblock(
            '''
            def func( with , some = 'Problems' ):


             syntax ='Ok'
             but = 'Its very messy'
             if None:
                    # syntax might not be perfect due to being cut off
                    ommiting_this_line_still_works=   True
            ''')
        >>> fixed_codeblock = ut.autofix_codeblock(codeblock)
        >>> print(fixed_codeblock)
    """
    # FIXME idk how to remove the blank line following the function with
    # autopep8. It seems to not be supported by them, but it looks bad.
    import autopep8
    arglist = ['--max-line-length', '80']
    if aggressive:
        arglist.extend(['-a'])
    if very_aggressive:
        arglist.extend(['-a', '-a'])
    if experimental:
        arglist.extend(['--experimental'])
    arglist.extend([''])
    autopep8_options = autopep8.parse_args(arglist)
    fixed_codeblock = autopep8.fix_code(codeblock, options=autopep8_options)
    return fixed_codeblock


def load_func_from_module(modname, funcname, verbose=True, moddir=None, modpath=None):
    r"""
    Args:
        modname (str):  module name
        funcname (str):  function name
        verbose (bool):  verbosity flag(Defaults to True)
        moddir (None): (Defaults to None)

    CommandLine:
        python -m utool.util_autogen load_func_from_module

    Example:
        >>> # DISABLE_DOCTEST
        >>> # UNSTABLE_DOCTEST
        >>> from utool.util_autogen import *  # NOQA
        >>> import utool as ut
        >>> #funcname = 'multi_plot'
        >>> modname = 'utool.util_path'
        >>> funcname = 'checkpath'
        >>> verbose = True
        >>> moddir = None
        >>> func, module, error_str = load_func_from_module(modname, funcname, verbose, moddir)
        >>> source = ut.get_func_sourcecode(func, strip_docstr=True, strip_comments=True)
        >>> keyname = ut.named_field('keyname', ut.REGEX_VARNAME)
        >>> default = ut.named_field('default', '[\'\"A-Za-z_][A-Za-z0-9_\'\"]*')
        >>> pattern = re.escape('kwargs.get(\'') + keyname + re.escape('\',')
        >>> kwarg_keys = [match.groupdict()['keyname'] for match in re.finditer(pattern, source)]
    """
    import utool as ut
    from os.path import join
    import imp
    func = None
    module = None
    error_str = None
    print('modname = %r' % (modname,))
    print('funcname = %r' % (funcname,))
    print('moddir = %r' % (moddir,))

    try:
        from xdoctest import static_analysis as static
        from xdoctest import utils
        print('modpath = {!r}'.format(modpath))
    except ImportError:
        modpath = None
    else:
        if modpath is None:
            modpath = static.modname_to_modpath(modname)
            if modpath is None:
                modname = modname.split('.')[-1]
                if moddir is not None:
                    modpath = join(moddir, modname + '.py')
                    if not exists(modpath):
                        modpath = None
                if modpath is None:
                    raise Exception('Cannot find modname={} in moddir={}'.format(modname, moddir))
        print('modpath = {!r}'.format(modpath))
        module = utils.import_module_from_path(modpath)
        print('module = {!r}'.format(module))
        try:
            func = eval('module.{}'.format(funcname))
        except AttributeError:
            imp.reload(module)
            func = eval('module.{}'.format(funcname))
        print('func = {!r}'.format(func))
        return func, module, error_str

    if not isinstance(modname, six.string_types):
        error_str = 'modname=%r is not a string. bad input' % (modname,)
    else:
        if False:
            # TODO: static analysis
            import jedi  # NOQA
            modpath = ut.util_import.get_modpath_from_modname(modname)
            script = jedi.Script(path=modpath)
            mod = script._get_module()
            # monkeypatch
            func.script = script
            func = None
            for name in mod.names_dict[funcname]:
                if name.parent.type == 'funcdef':
                    func = name.parent
                    break
            return func, mod, error_str
            # ut.get_modpath_from_modname(modname)
        if module is None:
            try:
                module = __import__(modname)
            except ImportError:
                if moddir is not None:
                    #parts =
                    # There can be a weird double import error thing happening here
                    # Rectify the dots in the filename
                    module = ut.import_module_from_fpath(join(moddir, modname.split('.')[-1] + '.py'))
                else:
                    raise
            #import inspect
            # try:
            #     imp.reload(module)
            # except Exception as ex:
            #     pass
        # if False:
        #     # Try removing pyc if it exists
        #     if module.__file__.endswith('.pyc'):
        #         ut.delete(module.__file__, verbose=False)
        #         try:
        #             module = __import__(modname)
        #         except ImportError:
        #             if moddir is not None:
        #                 module = ut.import_module_from_fpath(join(moddir, modname.split('.')[-1] + '.py'))
        #             else:
        #                 raise
        try:
            imp.reload(module)
        except Exception as ex:
            pass
        try:
            # FIXME: PYTHON 3
            execstr = ut.codeblock(
                '''
                try:
                    import {modname}
                    module = {modname}
                    #print('Trying to reload module=%r' % (module,))
                    imp.reload(module)
                except Exception:
                    # If it fails maybe the module is not in the path
                    if moddir is not None:
                        try:
                            import imp
                            import os
                            orig_dir = os.getcwd()
                            os.chdir(moddir)
                            modname_str = '{modname}'
                            modinfo = imp.find_module(modname_str, [moddir])
                            module = imp.load_module(modname_str, *modinfo)
                            #print('loaded module=%r' % (module,))
                        except Exception as ex:
                            ut.printex(ex, 'failed to imp.load_module')
                            pass
                        finally:
                            os.chdir(orig_dir)
                import imp
                import utool as ut
                imp.reload(ut.util_autogen)
                imp.reload(ut.util_inspect)
                try:
                    func = module.{funcname}
                except AttributeError:
                    docstr = 'Could not find attribute funcname={funcname} in modname={modname} This might be a reloading issue'
                    imp.reload(module)
                '''
            ).format(**locals())
            exec_locals = locals()
            exec_globals = globals()
            exec(execstr, exec_globals, exec_locals)
            func = exec_locals.get('func', None)
            module = exec_locals.get('module', None)
        except Exception as ex2:
            docstr = 'error ' + str(ex2)
            if verbose:
                import utool as ut
                #ut.printex(ex1, 'ex1')
                ut.printex(ex2, 'ex2', tb=True)
            testcmd = 'python -c "import utool; print(utool.auto_docstr(\'%s\', \'%s\'))"' % (modname, funcname)
            error_str = ut.formatex(ex2, 'ex2', tb=True, keys=['modname', 'funcname', 'testcmd'])
            error_str += '---' + execstr
    return func, module, error_str


def auto_docstr(modname, funcname, verbose=True, moddir=None, modpath=None, **kwargs):
    r"""
    called from vim. Uses strings of filename and modnames to build docstr

    Args:
        modname (str): name of a python module
        funcname (str): name of a function in the module

    Returns:
        str: docstr

    CommandLine:
        python -m utool.util_autogen auto_docstr
        python -m utool --tf auto_docstr

    Example:
        >>> # DISABLE_DOCTEST
        >>> import utool as ut
        >>> from utool.util_autogen import *  # NOQA
        >>> ut.util_autogen.rrr(verbose=False)
        >>> #docstr = ut.auto_docstr('ibeis.algo.hots.smk.smk_index', 'compute_negentropy_names')
        >>> modname = ut.get_argval('--modname', default='utool.util_autogen')
        >>> funcname = ut.get_argval('--funcname', default='auto_docstr')
        >>> moddir = ut.get_argval('--moddir', type_=str, default=None)
        >>> docstr = ut.util_autogen.auto_docstr(modname, funcname)
        >>> print(docstr)
    """
    #import utool as ut
    func, module, error_str = load_func_from_module(
        modname, funcname, verbose=verbose, moddir=moddir, modpath=modpath)
    if error_str is None:
        try:
            docstr = make_default_docstr(func, **kwargs)
        except Exception as ex:
            import utool as ut
            error_str = ut.formatex(ex, 'Caught Error in parsing docstr', tb=True)
            #ut.printex(ex)
            error_str += (
                '\n\nReplicateCommand:\n    '
                'python -m utool --tf auto_docstr '
                '--modname={modname} --funcname={funcname} --moddir={moddir}').format(
                    modname=modname, funcname=funcname, moddir=moddir)
            error_str += '\n kwargs='  + ut.repr4(kwargs)
            return error_str
    else:
        docstr = error_str
    return docstr


def print_auto_docstr(modname, funcname):
    """
    python -c "import utool; utool.print_auto_docstr('ibeis.algo.hots.smk.smk_index', 'compute_negentropy_names')"
    python -c "import utool;
    utool.print_auto_docstr('ibeis.algo.hots.smk.smk_index', 'compute_negentropy_names')"
    """
    docstr = auto_docstr(modname, funcname)
    print(docstr)


# <INVIDIAL DOCSTR COMPONENTS>

def make_args_docstr(argname_list, argtype_list, argdesc_list, ismethod,
                     va_name=None, kw_name=None, kw_keys=[]):
    r"""
    Builds the argument docstring

    Args:
        argname_list (list): names
        argtype_list (list): types
        argdesc_list (list): descriptions
        ismethod (bool): if generating docs for a method
        va_name (Optional[str]): varargs name
        kw_name (Optional[str]): kwargs name
        kw_keys (Optional[list]): accepted kwarg keys

    Returns:
        str: arg_docstr

    CommandLine:
        python -m utool.util_autogen make_args_docstr

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_autogen import *  # NOQA
        >>> argname_list = ['argname_list', 'argtype_list', 'argdesc_list']
        >>> argtype_list = ['list', 'list', 'list']
        >>> argdesc_list = ['names', 'types', 'descriptions']
        >>> va_name = 'args'
        >>> kw_name = 'kwargs'
        >>> kw_keys = ['']
        >>> ismethod = False
        >>> arg_docstr = make_args_docstr(argname_list, argtype_list,
        >>>                               argdesc_list, ismethod, va_name,
        >>>                               kw_name, kw_keys)
        >>> result = str(arg_docstr)
        >>> print(result)
        argname_list (list): names
        argtype_list (list): types
        argdesc_list (list): descriptions
        *args:
        **kwargs:

    """
    import utool as ut
    if ismethod:
        # Remove self from the list
        argname_list = argname_list[1:]
        argtype_list = argtype_list[1:]
        argdesc_list = argdesc_list[1:]

    argdoc_list = [arg + ' (%s): %s' % (_type, desc)
                   for arg, _type, desc in zip(argname_list, argtype_list, argdesc_list)]

    # Add in varargs and kwargs
    # References:
    # http://www.sphinx-doc.org/en/stable/ext/example_google.html#example-google
    if va_name is not None:
        argdoc_list.append('*' + va_name + ':')
    if kw_name is not None:
        import textwrap
        prefix = '**' + kw_name + ': '
        wrapped_lines = textwrap.wrap(', '.join(kw_keys), width=70 - len(prefix))
        sep = '\n' + (' ' * len(prefix))
        kw_keystr = sep.join(wrapped_lines)
        argdoc_list.append((prefix + kw_keystr).strip())

    # align?
    align_args = False
    if align_args:
        argdoc_aligned_list = ut.align_lines(argdoc_list, character='(')
        arg_docstr = '\n'.join(argdoc_aligned_list)
    else:
        arg_docstr = '\n'.join(argdoc_list)
    return arg_docstr


def make_returns_or_yeilds_docstr(return_type, return_name, return_desc):
    return_doctr = return_type + ': '
    if return_name is not None:
        return_doctr += return_name
        if len(return_desc) > 0:
            return_doctr += ' - '
    return_doctr += return_desc
    return return_doctr


def make_example_docstr(funcname=None, modname=None, argname_list=None,
                        defaults=None, return_type=None, return_name=None,
                        ismethod=False):
    """
    Creates skeleton code to build an example doctest

    Args:
        funcname (str):  function name
        modname (str):  module name
        argname_list (str):  list of argument names
        defaults (None):
        return_type (None):
        return_name (str):  return variable name
        ismethod (bool):

    Returns:
        str: examplecode

    CommandLine:
        python -m utool.util_autogen --test-make_example_docstr

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_autogen import *  # NOQA
        >>> funcname = 'make_example_docstr'
        >>> modname = 'utool.util_autogen'
        >>> argname_list = ['qaids', 'qreq_']
        >>> defaults = None
        >>> return_type = tuple
        >>> return_name = 'foo'
        >>> ismethod = False
        >>> examplecode = make_example_docstr(funcname, modname, argname_list, defaults, return_type, return_name, ismethod)
        >>> result = str(examplecode)
        >>> print(result)
        # DISABLE_DOCTEST
        from utool.util_autogen import *  # NOQA
        import utool as ut
        import ibeis
        species = ibeis.const.TEST_SPECIES.ZEB_PLAIN
        qaids = ibs.get_valid_aids(species=species)
        qreq_ = ibeis.testdata_qreq_()
        foo = make_example_docstr(qaids, qreq_)
        result = ('foo = %s' % (ut.repr2(foo),))
        print(result)
    """
    import utool as ut

    examplecode_lines = []
    top_import_fmstr = 'from {modname} import *  # NOQA'
    top_import = top_import_fmstr.format(modname=modname)
    import_lines = [top_import]
    if modname.startswith('utool'):
        import_lines += ['import utool as ut']
    # is_show_func = not modname.startswith('utool') and not modname.startswith('mtgmonte')
    is_show_func = modname.startswith('plottool')

    # TODO: Externally register these
    default_argval_map = {
        'ibs'       : 'ibeis.opendb(defaultdb=\'testdb1\')',
        'testres'   : 'ibeis.testdata_expts(\'PZ_MTEST\')',
        'qreq_'     : 'ibeis.testdata_qreq_()',
        'cm_list'   : 'qreq_.execute()',
        'cm'        : 'qreq_.execute()[0]',
        'aid_list'  : 'ibs.get_valid_aids()',
        'nid_list'  : 'ibs._get_all_known_nids()',
        'qaids'     : 'ibs.get_valid_aids(species=species)',
        'daids'     : 'ibs.get_valid_aids(species=species)',
        'species'   : 'ibeis.const.TEST_SPECIES.ZEB_PLAIN',
        'kpts'      : 'vt.dummy.get_dummy_kpts()',
        'dodraw'    : 'ut.show_was_requested()',
        'img_fpath' : 'ut.grab_test_imgpath(\'carl.jpg\')',
        'gfpath'    : 'ut.grab_test_imgpath(\'carl.jpg\')',
        'img'       : 'vt.imread(img_fpath)',
        'img_in'    : 'vt.imread(img_fpath)',
        'bbox'      : '(10, 10, 50, 50)',
        'theta'     : '0.0',
        'rng'       : 'np.random.RandomState(0)',
    }
    import_depends_map = {
        'ibeis':    'import ibeis',
        'vt':       'import vtool as vt',
        #'img':      'import vtool as vt',  # TODO: remove. fix dependency
        #'species':  'import ibeis',
    }
    var_depends_map = {
        'species':   ['ibeis'],
        'ibs':       ['ibeis'],
        'testres': ['ibeis'],
        'kpts':      ['vt'],
        #'qreq_':     ['ibs', 'species', 'daids', 'qaids'],
        'qreq_':     ['ibeis'],
        'qaids':     ['ibs'],
        'daids':     ['ibs'],
        'qaids':     ['species'],
        'daids':     ['species'],
        'img':       ['img_fpath', 'vt'],
    }

    def find_arg_defaultrepr(argname, val):
        import types
        if val == '?':
            if argname in default_argval_map:
                val = ut.PythonStatement(default_argval_map[argname])
                if argname in import_depends_map:
                    import_lines.append(import_depends_map[argname])
        elif isinstance(val, types.ModuleType):
            return val.__name__
        return repr(val)

    # augment argname list with dependencies
    dependant_argnames = []  # deque()
    def append_dependant_argnames(argnames, dependant_argnames):
        """ use hints to add known dependencies for certain argument inputs """
        for argname in argnames:
            # Check if argname just implies an import
            if argname in import_depends_map:
                import_lines.append(import_depends_map[argname])
            # Check if argname was already added as dependency
            if (argname not in dependant_argnames and argname not in
                 argname_list and argname not in import_depends_map):
                dependant_argnames.append(argname)
            # Check if argname has dependants
            if argname in var_depends_map:
                argdeps = var_depends_map[argname]
                # RECURSIVE CALL
                append_dependant_argnames(argdeps, dependant_argnames)
    append_dependant_argnames(argname_list, dependant_argnames)

    # Define argnames and dependencies in example code
    # argnames prefixed with dependeancies
    argname_list_ = list(dependant_argnames) + argname_list

    # Default example values
    defaults_ = [] if defaults is None else defaults
    num_unknown = (len(argname_list_) - len(defaults_))
    default_vals = ['?'] * num_unknown + list(defaults_)
    arg_val_iter = zip(argname_list_, default_vals)
    inferred_defaults = [find_arg_defaultrepr(argname, val)
                         for argname, val in arg_val_iter]
    argdef_lines = ['%s = %s' % (argname, inferrepr)
                    for argname, inferrepr in
                    zip(argname_list_, inferred_defaults)]
    import_lines = ut.unique_ordered(import_lines)

    if any([inferrepr == repr('?') for inferrepr in inferred_defaults]):
        examplecode_lines.append('# DISABLE_DOCTEST')
    else:
        # Enable the test if it can be run immediately
        examplecode_lines.append('# DISABLE_DOCTEST')

    examplecode_lines.extend(import_lines)
    examplecode_lines.extend(argdef_lines)
    # Default example result assignment
    result_assign = ''
    result_print = None
    if 'return_name' in vars():
        if return_type is not None:
            if return_name is None:
                return_name = 'result'
            result_assign = return_name + ' = '
            result_print = 'print(result)'  # + return_name + ')'
    # Default example call
    if ismethod:
        selfname = argname_list[0]
        methodargs = ', '.join(argname_list[1:])
        tup = (selfname, '.', funcname, '(', methodargs, ')')
        example_call = ''.join(tup)
    else:
        funcargs = ', '.join(argname_list)
        tup = (funcname, '(', funcargs, ')')
        example_call = ''.join(tup)
    # Append call line
    examplecode_lines.append(result_assign + example_call)
    if result_print is not None:
        if return_name != 'result':
            #examplecode_lines.append('result = str(' + return_name + ')')
            result_line_fmt = 'result = (\'{return_name} = %s\' % (ut.repr2({return_name}),))'
            result_line = result_line_fmt.format(return_name=return_name)
            examplecode_lines.append(result_line)
        examplecode_lines.append(result_print)

    # TODO: infer this
    if is_show_func:
        examplecode_lines += [
            '# xdoctest: +REQUIRES(--show)',
        ]

    examplecode = '\n'.join(examplecode_lines)
    return examplecode


def make_cmdline_docstr(funcname, modname):
    #cmdline_fmtstr = 'python -m {modname} --test-{funcname}'  # --enableall'
    #cmdline_fmtstr = 'python -m {modname} --exec-{funcname}'  # --enableall'
    if False and  '.' in modname and '.' not in funcname:
        pkg = modname.split('.')[0]
        # TODO check if __main__ exists with the necessary utool stuffs
        # TODO check if --show should be given
        cmdline_fmtstr = 'python -m {pkg} --tf {funcname}'  # --enableall'
        return cmdline_fmtstr.format(**locals())
    else:
        # TODO: infer this
        is_show_func = 'draw' in funcname or 'show' in funcname
        if is_show_func:
            cmdline_fmtstr = 'python -m {modname} {funcname} --show'
        else:
            cmdline_fmtstr = 'python -m {modname} {funcname}'  # --enableall'
        return cmdline_fmtstr.format(**locals())

# </INVIDIAL DOCSTR COMPONENTS>


def make_docstr_block(header, block):
    import utool as ut
    indented_block = '\n' + ut.indent(block)
    docstr_block = ''.join([header, ':', indented_block])
    return docstr_block


def make_default_docstr(func, with_args=True, with_ret=True,
                        with_commandline=True, with_example=True,
                        with_header=False, with_debug=False):
    r"""
    Tries to make a sensible default docstr so the user
    can fill things in without typing too much

    # TODO: Interleave old documentation with new documentation

    Args:
        func (function): live python function
        with_args (bool):
        with_ret (bool): (Defaults to True)
        with_commandline (bool): (Defaults to True)
        with_example (bool): (Defaults to True)
        with_header (bool): (Defaults to False)
        with_debug (bool): (Defaults to False)

    Returns:
        tuple: (argname, val)

    Ignore:
        pass

    CommandLine:
        python -m utool.util_autogen --exec-make_default_docstr --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_autogen import *  # NOQA
        >>> import utool as ut
        >>> func = ut.make_default_docstr
        >>> #func = ut.make_args_docstr
        >>> #func = PythonStatement
        >>> func = auto_docstr
        >>> default_docstr = make_default_docstr(func)
        >>> result = str(default_docstr)
        >>> print(result)

    """
    import utool as ut
    #from utool import util_inspect
    funcinfo = ut.util_inspect.infer_function_info(func)

    argname_list   = funcinfo.argname_list
    argtype_list   = funcinfo.argtype_list
    argdesc_list   = funcinfo.argdesc_list
    return_header  = funcinfo.return_header
    return_type    = funcinfo.return_type
    return_name    = funcinfo.return_name
    return_desc    = funcinfo.return_desc
    funcname       = funcinfo.funcname
    modname        = funcinfo.modname
    defaults       = funcinfo.defaults
    num_indent     = funcinfo.num_indent
    needs_surround = funcinfo.needs_surround
    funcname       = funcinfo.funcname
    ismethod       = funcinfo.ismethod
    va_name        = funcinfo.va_name
    kw_name        = funcinfo.kw_name
    kw_keys        = funcinfo.kw_keys

    docstr_parts = []
    # Header part
    if with_header:
        header_block = funcname
        docstr_parts.append(header_block)

    # Args part
    if with_args and len(argname_list) > 0:
        argheader = 'Args'
        arg_docstr = make_args_docstr(argname_list, argtype_list, argdesc_list,
                                      ismethod, va_name, kw_name, kw_keys)
        argsblock = make_docstr_block(argheader, arg_docstr)

        docstr_parts.append(argsblock)

    # if False:
    #     with_kw = with_args
    #     if with_kw and len(kwarg_keys) > 0:
    #         #ut.embed()
    #         import textwrap
    #         kwargs_docstr = ', '.join(kwarg_keys)
    #         kwargs_docstr = '\n'.join(textwrap.wrap(kwargs_docstr))
    #         kwargsblock = make_docstr_block('Kwargs', kwargs_docstr)
    #         docstr_parts.append(kwargsblock)

    # Return / Yeild part
    if with_ret and return_header is not None:
        if return_header is not None:
            return_doctr = make_returns_or_yeilds_docstr(return_type, return_name, return_desc)
            returnblock = make_docstr_block(return_header, return_doctr)
            docstr_parts.append(returnblock)

    # Example part
    # try to generate a simple and unit testable example
    if with_commandline:
        cmdlineheader = 'CommandLine'
        cmdlinecode = make_cmdline_docstr(funcname, modname)
        cmdlineblock = make_docstr_block(cmdlineheader, cmdlinecode)
        docstr_parts.append(cmdlineblock)

    if with_example:
        exampleheader = 'Example'
        examplecode = make_example_docstr(funcname, modname, argname_list,
                                          defaults, return_type, return_name,
                                          ismethod)
        examplecode_ = ut.indent(examplecode, '>>> ')
        exampleblock = make_docstr_block(exampleheader, examplecode_)
        docstr_parts.append(exampleblock)

    # DEBUG part (in case something goes wrong)
    if with_debug:
        debugheader = 'Debug'
        debugblock = ut.codeblock(
            '''
            num_indent = {num_indent}
            '''
        ).format(num_indent=num_indent)
        debugblock = make_docstr_block(debugheader, debugblock)
        docstr_parts.append(debugblock)

    # Enclosure / Indentation Parts
    if needs_surround:
        docstr_parts = ['r"""'] + ['\n\n'.join(docstr_parts)] + ['"""']
        default_docstr = '\n'.join(docstr_parts)
    else:
        default_docstr = '\n\n'.join(docstr_parts)

    docstr_indent = ' ' * (num_indent + 4)
    default_docstr = ut.indent(default_docstr, docstr_indent)
    return default_docstr


def remove_codeblock_syntax_sentinals(code_text):
    r"""
    Removes template comments and vim sentinals

    Args:
        code_text (str):

    Returns:
        str: code_text_
    """
    flags = re.MULTILINE | re.DOTALL
    code_text_ = code_text
    code_text_ = re.sub(r'^ *# *REM [^\n]*$\n?', '', code_text_, flags=flags)
    code_text_ = re.sub(r'^ *# STARTBLOCK *$\n', '', code_text_, flags=flags)
    code_text_ = re.sub(r'^ *# ENDBLOCK *$\n?', '', code_text_, flags=flags)
    code_text_ = code_text_.rstrip()
    return code_text_


def make_default_module_maintest(modname, modpath=None, test_code=None):
    """
    make_default_module_maintest

    DEPRICATE

    TODO: use path relative to home dir if the file is a script

    Args:
        modname (str):  module name

    Returns:
        str: text source code

    CommandLine:
        python -m utool.util_autogen --test-make_default_module_maintest

    References:
        http://legacy.python.org/dev/peps/pep-0338/

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_autogen import *  # NOQA
        >>> modname = 'utool.util_autogen'
        >>> text = make_default_module_maintest(modname)
        >>> result = str(text)
        >>> print(result)
    """
    import utool as ut
    # Need to use python -m to run a module
    # otherwise their could be odd platform specific errors.
    #python -c "import utool, {modname};
    # ut.doctest_funcs({modname}, allexamples=True)"
    #in_pythonpath, module_type, path = find_modname_in_pythonpath(modname)
    # only use the -m if it is part of a package directory

    if modpath is not None:
        moddir = dirname(modpath)
        pkginit_fpath = join(moddir, '__init__.py')
        use_modrun = exists(pkginit_fpath)
    else:
        use_modrun = True

    if use_modrun:
        pyargs = '-m ' + modname
    else:
        if ut.WIN32:
            modpath = normpath(modpath).replace(expanduser('~'), '%HOME%')
            pyargs = '-B ' + ut.ensure_unixslash(modpath)
        else:
            modpath = normpath(modpath).replace(expanduser('~'), '~')
            pyargs = modpath

    cmdline = ut.codeblock(
        '''
        python {pyargs}
        ''')

    if not use_modrun:
        if ut.WIN32:
            augpath = 'set PYTHONPATH=%PYTHONPATH%' + os.pathsep + moddir
        else:
            augpath = 'export PYTHONPATH=$PYTHONPATH' + os.pathsep + moddir
        cmdline = augpath + '\n' + cmdline

    cmdline = ut.indent(cmdline, ' ' * 8).lstrip(' ').format(pyargs=pyargs)

    if test_code is None:
        test_code = ut.codeblock(
            r'''
            import multiprocessing
            multiprocessing.freeze_support()  # for win32
            import utool as ut  # NOQA
            ut.doctest_funcs()
            ''')

    test_code = ut.indent(test_code, ' ' * 4).lstrip(' ')

    text = ut.codeblock(
        r'''
        # STARTBLOCK
        if __name__ == '__main__':
            r"""
            CommandLine:
                {cmdline}
            """
            {test_code}
        # ENDBLOCK
        '''
    ).format(cmdline=cmdline, test_code=test_code)
    print('test_code = {!r}'.format(test_code))
    text = remove_codeblock_syntax_sentinals(text)
    return text


def is_modname_in_pythonpath(modname):
    in_pythonpath, module_type, path = find_modname_in_pythonpath(modname)
    print(module_type)
    return in_pythonpath


def find_modname_in_pythonpath(modname):
    import sys
    from os.path import exists, join
    rel_modpath = modname.replace('.', '/')
    in_pythonpath = False
    module_type = None
    path_list = sys.path
    path_list = os.environ['PATH'].split(os.pathsep)
    for path in path_list:
        full_modpath = join(path, rel_modpath)
        if exists(full_modpath + '.py'):
            in_pythonpath = True
            module_type = 'module'
            break
        if exists(join(full_modpath, '__init__.py')):
            in_pythonpath = True
            module_type = 'package'
            break
    return in_pythonpath, module_type, path

    #ut.get_modpath(modname)
    #import imp
    #tup = imp.find_module(modname)
    #(file, filename, (suffix, mode, type))


if __name__ == '__main__':
    """
    CommandLine:
        python ibeis/control/template_generator.py --tbls annotations --Tflags getters native
        python -c "import utool, utool.util_autogen; utool.doctest_funcs(utool.util_autogen, allexamples=True)"
        python -m utool.util_autogen
        python -m utool.util_autogen --allexamples
        python -m utool.util_autogen --allexamples --noface --nosrc
    """
    import multiprocessing
    multiprocessing.freeze_support()  # for win32
    import utool as ut  # NOQA
    ut.doctest_funcs()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function
import functools
from utool import util_inject
print, rrr, profile = util_inject.inject2(__name__)


def compose_functions(*func_list):
    """
    Referenes:
        https://mathieularose.com/function-composition-in-python/
    """
    def apply_composition(f, g):
        def compose(x):
            return f(g(x))
        return compose
    composed_func = functools.reduce(apply_composition, func_list)
    return composed_func


def identity(input_):
    """ identity function """
    return input_

# DEPRICATE EVERYTHING BELOW HERE


#def uinput_1to1(func, input_):
#    """ universal input (really just accept list or tuple as input to a list
#    only function)

#    Move to guitool
#    """
#    if isinstance(input_, (tuple, list)):
#        output_ = list(map(func, input_))
#    else:
#        output_ = func(input_)
#    return output_


#def general_get(getter, index, **kwargs):
#    """ Works with getter funcs or indexable read/write arrays """
#    if hasattr(getter, '__getitem__'):
#        val = getter[index]
#    else:
#        val = getter(index, **kwargs)
#    return val


#def general_set(setter, index, val, **kwargs):
#    """ Works with setter funcs or indexable read/write arrays """
#    if hasattr(setter, '__setitem__'):
#        setter[index] = val
#    else:
#        setter(index, val, **kwargs)



# -*- coding: utf-8 -*-
lazy_module_attrs =  ['_modname', '_module', '_load_module']


class LazyModule(object):
    """
    Waits to import the module until it is actually used

    """
    def __init__(self, modname):
        self._modname = modname
        self._module = None

    def _load_module(self):
        #print('loading module')
        self._module =  __import__(self._modname, globals(), locals(), fromlist=[], level=0)

    def __str__(self):
        return 'LazyModule(%s)' % (self._modname,)

    def __dir__(self):
        self._load_module()
        return dir(self._module)

    def __getattr__(self, item):
        """Maps values to attributes.
        Only called if there *isn't* an attribute with this name
        """
        if item in lazy_module_attrs:
            return super(LazyModule, self).__getattr__(item)
        self._load_module()
        return getattr(self._module, item)

    def __setattr__(self, item, value):
        """Maps attributes to values.
        Only if we are initialised
        """
        if item in lazy_module_attrs:
            return super(LazyModule, self).__setattr__(item, value)
        self._load_module()
        setattr(self._module, item, value)


modname = 'theano'
theano = LazyModule(modname)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
downloads an archive file and then unzips it into a directory with the same name
as the archive (sans the .zip or .tar.gz)

grabzippedurl.py "https://download.zotero.org/standalone/4.0.26.3/Zotero-4.0.26.3_linux-x86_64.tar.bz2"

"""
from __future__ import absolute_import, division, print_function
import sys
import utool

if __name__ == '__main__':
    url = sys.argv[1]
    download_dir = '.'
    utool.grab_zipped_url(url, download_dir=download_dir, cleanup=False, spoof=True)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
import six
code = '''
import sys

IGNORE_TRACEBACK = not ('--nosmalltb' in sys.argv or '--noignoretb' in sys.argv)  # FIXME: dupliated in util_decor


# Module for funcs that need python 2 syntax to work in python 2
def ignores_exc_tb(*args, **kwargs):
    r"""
    PYTHON 2 ONLY VERSION -- needs to be in its own file for syntactic reasons

    ignore_exc_tb decorates a function and remove both itself
    and the function from any exception traceback that occurs.

    This is useful to decorate other trivial decorators
    which are polluting your stacktrace.

    if IGNORE_TRACEBACK is False then this decorator does nothing
    (and it should do nothing in production code!)

    References:
        https://github.com/jcrocholl/pep8/issues/34  # NOQA
        http://legacy.python.org/dev/peps/pep-3109/
    """
    outer_wrapper = kwargs.get('outer_wrapper', True)
    def ignores_exc_tb_closure(func):
        if not IGNORE_TRACEBACK:
            # if the global enforces that we should not ignore anytracebacks
            # then just return the original function without any modifcation
            return func
        from utool import util_decor
        #@wraps(func)
        def wrp_noexectb(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception:
                # Define function to reraise with python 2 syntax
                #exc_type, exc_value, exc_traceback = sys.exc_info()
                # Code to remove this decorator from traceback
                # Remove two levels to remove this one as well
                exc_type, exc_value, exc_traceback = sys.exc_info()
                try:
                    exc_traceback = exc_traceback.tb_next
                    exc_traceback = exc_traceback.tb_next
                    #exc_traceback = exc_traceback.tb_next
                except Exception:
                    print('too many reraise')
                    pass
                raise exc_type, exc_value, exc_traceback
        if outer_wrapper:
            wrp_noexectb = util_decor.preserve_sig(wrp_noexectb, func)
        return wrp_noexectb
    if len(args) == 1:
        # called with one arg means its a function call
        func = args[0]
        return ignores_exc_tb_closure(func)
    else:
        # called with no args means kwargs as specified
        return ignores_exc_tb_closure
'''


if six.PY2:
    exec(code, globals(), globals())



"""This is a minimal Python client for Mads Haahr's random number generator at www.random.org

# This tiny set of functions only implements a subset of the HTTP interface available. In particular it only uses the 'live'
# random number generator, and doesn't offer the option of using the alternative 'stored' random
# number sets. However, it should be obvious how to extend it by sending requests with different parameters.
# The web service code is modelled on Mark Pilgrim's "Dive into Python" tutorial at http://www.diveintopython.org/http_web_services
# This client by George Dunbar, University of Warwick (Copyright George Dunbar, 2008)
# It is distributed under the Gnu General Public License.
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    See <http://www.gnu.org/licenses/> for a copy of the GNU General Public License.
    For use that falls outside this license, please contact me.


To use in a python script or at the interactive prompt
(randomwrapy.py has to be in the Python search path, of course):

from randomwrapy import *

rnumlistwithoutreplacement(0, 12)
   # returns a list of the integers 0 - 12 inclusive, in a random order

rnumlistwithreplacement(12, 5)
   # returns 12 integers from the range [0, 5]

rnumlistwithreplacement(12, 5, 2)
   # returns 12 integers from the range [2, 5]

rrandom()
   # returns a random float in the range [0, 1]

reportquota()
   # tells you how many bits you have available; visit www.random.org/quota for more information

Arguments where given are (must be) numbers, of course.
There is almost no error checking in these scripts! For example, if
the web site is down, Python will simply raise an exception and report the
http error code. See worldrandom.py for an alternative implementation
that goes a little further with error checking.

"""

from six.moves import urllib


def rnumlistwithoutreplacement(min, max):
    """Returns a randomly ordered list of the integers between min and max"""
    if checkquota() < 1:
        raise Exception("Your www.random.org quota has already run out.")
    requestparam = build_request_parameterNR(min, max)
    request = urllib.request.Request(requestparam)
    request.add_header('User-Agent', 'randomwrapy/0.1 very alpha')
    opener = urllib.request.build_opener()
    numlist = opener.open(request).read()
    return numlist.split()

# helper


def build_request_parameterNR(min, max):
    randomorg = 'http://www.random.org/sequences/?min='
    vanilla = '&format=plain&rnd=new'
    params = str(min) + '&max=' + str(max)
    return randomorg + params + vanilla


def rnumlistwithreplacement(howmany, max, min=0):
    """Returns a list of howmany integers with a maximum value = max.
    The minimum value defaults to zero."""
    if checkquota() < 1:
        raise Exception("Your www.random.org quota has already run out.")
    requestparam = build_request_parameterWR(howmany, min, max)
    request = urllib.request.Request(requestparam)
    request.add_header('User-Agent', 'randomwrapy/0.1 very alpha')
    opener = urllib.request.build_opener()
    numlist = opener.open(request).read()
    return numlist.split()

"""
Example usage:

Roll a dice 12 times (returning integers in the range [0,5]):
  rnumlistwithreplacement(12, 5)

Roll a dice 12 times (returning integers in the more familiar range [1,6]):
  rnumlistwithreplacement(12, 6, 1)
"""

# helper


def build_request_parameterWR(howmany, min, max):
    randomorg = 'http://www.random.org/integers/?num='
    vanilla = '&col=1&base=10&format=plain&rnd=new'
    params = str(howmany) + '&min=' + str(min) + '&max=' + str(max)
    return randomorg + params + vanilla

# next function is prototype for integration with random module of python
# see worldrandom module for a more developed implementation


def rrandom():
    """Get the next random number in the range [0.0, 1.0].
    Returns a float."""
    import urllib.request
    import urllib.error
    import urllib.parse
    if checkquota() < 1:
        raise Exception("Your www.random.org quota has already run out.")
    request = urllib.request.Request(
        'http://www.random.org/integers/?num=1&min=0&max=1000000000&col=1&base=10&format=plain&rnd=new')
    request.add_header('User-Agent', 'randomwrapy/0.1 very alpha')
    opener = urllib.request.build_opener()
    numlist = opener.open(request).read()
    num = numlist.split()[0]
    return float(num) / 1000000000


def checkquota():
    request = urllib.request.Request(
        "http://www.random.org/quota/?format=plain")
    request.add_header('User-Agent', 'randomwrapy/0.1 very alpha')
    opener = urllib.request.build_opener()
    quota = opener.open(request).read()
    return int(quota)


def reportquota():
    request = urllib.request.Request(
        "http://www.random.org/quota/?format=plain")
    request.add_header('User-Agent', 'randomwrapy/0.1 very alpha')
    opener = urllib.request.build_opener()
    quota = opener.open(request).read()
    print("This IP address has", quota,
          "bits left. Visit http://www.random.org/quota for more information.")



#!/usr/bin/env python
"""
TODO: export from utool

        python -m utool.util_inspect check_module_usage --pat="util_git.py"
"""
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function
import sys
import os
import re
import six
from six.moves import zip
from os.path import exists, join, dirname, isdir, basename
from utool import util_dev
from utool import util_inject
from utool import util_class
from utool import util_path
from utool import util_decor
from utool import util_list
print, rrr, profile = util_inject.inject2(__name__)


def _syscmd(cmdstr):
    print('RUN> ' + cmdstr)
    os.system(cmdstr)


def _cd(dir_):
    dir_ = util_path.truepath(dir_)
    print('> cd ' + dir_)
    os.chdir(dir_)


@six.add_metaclass(util_class.ReloadingMetaclass)
class RepoManager(util_dev.NiceRepr):
    """
    Batch git operations on multiple repos
    """
    def __init__(rman, repo_urls=None, code_dir=None, userid=None,
                 permitted_repos=None, label='', pythoncmd=None):
        if userid is None:
            userid = None
        if permitted_repos is None:
            permitted_repos = []
        rman.permitted_repos = permitted_repos
        rman.code_dir = code_dir
        rman.userid = userid
        rman.repos = []
        rman.label = label
        rman.pythoncmd = pythoncmd

        if repo_urls is not None:
            rman.add_repos(repo_urls)

    def union(rman, other):
        new = RepoManager()
        new.userid = rman.userid
        new.code_dir = rman.code_dir
        new.permitted_repos = rman.permitted_repos + other.permitted_repos
        new.repos = rman.repos + other.repos
        return new

    @property
    def repo_urls(rman):
        return [repo.url for repo in rman.repos]

    @property
    def repo_dirs(rman):
        return [repo.url for repo in rman.repos]

    def __nice__(rman):
        return '(num=%d)' % (len(rman.repo_urls))

    def __getitem__(rman, name):
        for repo in rman.repos:
            if name in repo.aliases:
                return repo
        raise KeyError(name)

    def ensure(rman):
        print('Ensuring that respos are checked out')
        for repo in rman.repos:
            if repo.url is not None:
                repo.clone()

    def add_repo(rman, repo):
        repo._fix_url(rman.userid, rman.permitted_repos)
        rman.repos.append(repo)

    def add_repos(rman, repo_urls=None, code_dir=None):
        if code_dir is None:
            code_dir = rman.code_dir
        assert code_dir is not None, 'Must specify the checkout code_dir'
        repos = [Repo(url, code_dir, pythoncmd=rman.pythoncmd)
                 for url in repo_urls]
        for repo in repos:
            repo._fix_url(rman.userid, rman.permitted_repos)
        rman.repos.extend(repos)

    def issue(rman, command, sudo=False):
        """ Runs a command on all of managed repos """
        print('+------- GG_COMMAND -------')
        print('| sudo=%s' % sudo)
        print('| command=%s' % command)
        for repo in rman.repos:
            if exists(repo.dpath):
                repo.issue(command, sudo=sudo)
            else:
                print('Repo %r not found' % (repo,))
        print('L___ FINISHED GG_COMMAND ___')

    def check_importable(rman):
        import utool as ut
        label = ' %s' % rman.label if rman.label else rman.label
        missing = []
        print('Checking if%s modules are importable' % (label,))
        msg_list = []
        recommended_fixes = []
        for repo in rman.repos:
            flag, msg, errors = repo.check_importable()
            if not flag:
                msg_list.append('  * !!!%s REPO %s HAS IMPORT ISSUES' % (label.upper(), repo,))
                if any([str(ex).find('undefined symbol') > -1 for ex in errors]):
                    recommended_fixes.append('rebuild')
                else:
                    recommended_fixes.append(None)
                if ut.VERBOSE:
                    msg_list.append(ut.indent(msg, '    '))
                missing.append(repo)
            else:
                if ut.VERBOSE:
                    msg_list.append(ut.indent(msg, '    '))
        print('\n'.join(msg_list))
        problems = list(zip(missing, recommended_fixes))
        return problems

    def check_installed(rman):
        import utool as ut
        label = ' %s' % rman.label if rman.label else rman.label
        missing = []
        msg_list = []
        print('Checking if%s modules are installed' % (label,))
        for repo in rman.repos:
            flag, msg = repo.check_installed()
            if not flag:
                msg_list.append('  * !!!%s REPO %s NEEDS TO BE INSTALLED' % (label.upper(), repo,))
                if ut.VERBOSE:
                    msg_list.append(ut.indent(msg, '    '))
                missing.append(repo)
            # else:
            #     print('  * found%s module = %s' % (label, repo,))
        print('\n'.join(msg_list))
        return missing

    def check_cpp_build(rman):
        import utool as ut
        label = ' %s' % rman.label if rman.label else rman.label
        missing = []
        print('Checking if%s modules are built' % (label,))
        for repo in rman.repos:
            flag, msg = repo.check_cpp_build()
            if not flag:
                print('  * !!!%s REPO %s NEEDS TO BE BUILT' % (label.upper(), repo,))
                if ut.VERBOSE:
                    print(ut.indent(msg, '    '))
                missing.append(repo)
        return missing

    def custom_build(rman):
        print('Custom Build')
        for repo in rman.repos:
            script = repo.get_script('build')
            if script is not None:
                script.exec_()

    def custom_install(rman):
        print('Custom Install')
        for repo in rman.repos:
            script = repo.get_script('install')
            if script is not None:
                script.exec_()

    def only_with_pysetup(rman):
        rman2 = RepoManager()
        rman2.code_dir = rman.code_dir
        rman2.permitted_repos = rman.permitted_repos
        rman2.code_dir = rman.code_dir
        rman2.userid = rman.userid
        rman2.label = rman.label
        rman2.repos = [repo for repo in rman.repos if repo.dpath and exists(join(repo.dpath, 'setup.py'))]
        return rman2


@six.add_metaclass(util_class.ReloadingMetaclass)
class Repo(util_dev.NiceRepr):
    """
    Handles a Python module repository
    """
    def __init__(repo, url=None, code_dir=None, dpath=None,
                 modname=None, pythoncmd=None):
        # modname might need to be called egg?
        import utool as ut
        if url is not None and '.git@' in url:
            # parse out specific branch
            repo.default_branch = url.split('@')[-1]
            url = '@'.join(url.split('@')[:-1])
        else:
            repo.default_branch = None
        repo.url = url
        repo._modname = None
        if modname is None:
            modname = []
        repo._modname_hints = ut.ensure_iterable(modname)
        repo.dpath = None
        repo.scripts = {}
        if pythoncmd is None:
            import sys
            pythoncmd = sys.executable
        repo.pythoncmd = pythoncmd

        if dpath is None and repo.url is not None and code_dir is not None:
            dpath = join(code_dir, repo.reponame)
        if dpath is not None:
            repo.dpath = util_path.unixpath(dpath)

    def infer_info(repo):
        if repo.url is None:
            repo.url = list(repo.as_gitpython().remotes[0].urls)[0]

    # --- GIT PYTHON STUFF ---

    @util_decor.memoize
    def as_gitpython(repo):
        """ pip install gitpython """
        import git
        gitrepo = git.Repo(repo.dpath)
        return gitrepo

    @property
    def remotes(repo):
        remotes = repo.as_gitpython().remotes
        remote_dict = {}
        for remote in remotes:
            pass
            remote_info = repo._remote_info(remote)
            if remote_info is not None:
                name = remote_info.pop('name')
                remote_dict[name] = remote_info
        return remote_dict

    def _remote_info(repo, remote):
        OLD = False
        if OLD:
            remote_details = remote.repo.git.remote("get-url", remote.name, '--push')
            # TODO push into gitpython
            urls = [line for line in remote_details.split('\n')]
        else:
            urls = list(remote.urls)

        # urls = list(remote.urls)
        if len(urls) == 0:
            print('[git] WARNING: repo %r has no remote urls' % (repo,))
            remote_info = None
        else:
            if len(urls) > 1:
                print('[git] WARNING: repo %r has multiple urls' % (repo,))
            url = urls[0]
            url = url.replace('github.com:/', 'github.com:')
            remote_info = {}
            url_parts = re.split('[@/:]', url)
            # TODO: parse what format the url is in, ssh/http/https
            idx = util_list.listfind(url_parts, 'github.com')
            remote_info['name'] = remote.name
            remote_info['url'] = url
            if idx is not None:
                username = url_parts[idx + 1]
                remote_info['host'] = 'github'
                remote_info['username'] = username
        return remote_info

    def _ensure_remote_exists(repo, remote_name, remote_url, fmt=None):
        # Remote the remote if it is not in the correct format
        if fmt is None:
            if remote_url.startswith('git@'):
                fmt = 'ssh'
            elif remote_url.startswith('https://'):
                fmt = 'https'
            else:
                raise ValueError('bad format')
        gitrepo = repo.as_gitpython()
        remotes = repo.remotes
        incorrect_version = False
        if remote_url in remotes:
            # Check correct version (SSH or HTTPS)
            wildme_remote_ = remotes[remote_url]
            wildme_url_ = wildme_remote_['url']
            is_ssh = '@' in wildme_url_
            incorrect_version = (is_ssh and fmt == 'https') or (not is_ssh and fmt == 'ssh')
            if incorrect_version:
                print('  * Deleting bad version remote %r: %r' % (remote_name, remote_url))
                gitrepo.delete_remote(remote_name)

        # Ensure there is a remote under the wildme name
        if remote_name not in repo.remotes or incorrect_version:
            print('  * Create remote %r: %r' % (remote_name, remote_url))
            gitrepo.create_remote(remote_name, remote_url)
        return incorrect_version

    def _new_remote_url(repo, host=None, user=None, reponame=None, fmt=None):
        import utool as ut
        if reponame is None:
            reponame = repo.reponame
        if host is None:
            host = 'github.com'
        if fmt is None:
            fmt = 'ssh'
        if host == 'github.com':
            assert user is not None, 'github needs a user'
        url_fmts = {
            'https': ('https://', '/'),
            'ssh':   ('git@', ':'),
        }
        prefix, sep = url_fmts[fmt]
        user_ = '' if user is None else user + '/'
        parts = [prefix, host, sep, user_, reponame, '.git']
        parts = ut.filter_Nones(parts)
        url = ''.join(parts)
        return url

    def reset_branch_to_remote(repo, branch, hard=True):
        """
        does a git reset --hard to whatever remote the branch is assigned to
        """
        remote = repo.get_branch_remote(branch)
        kw = dict(remote=remote, branch=branch)
        if hard:
            kw['flags'] = '--hard'
        repo.issue('git reset {flags} {remote}/{branch}'.format(**kw))

    def get_branch_remote(repo, branch):
        gitrepo = repo.as_gitpython()
        gitbranch = gitrepo.branches[branch]
        remote = gitbranch.tracking_branch().remote_name
        return remote

    def set_branch_remote(repo, branch, remote, remote_branch=None):
        if remote_branch is None:
            remote_branch = branch
        fmt = 'git branch --set-upstream-to={remote}/{remote_branch} {branch} '
        cmd = fmt.format(branch=branch, remote=remote,
                         remote_branch=remote_branch)
        repo.issue(cmd)

    @property
    def active_branch(repo):
        return repo.as_gitpython().active_branch.name

    @property
    def active_remote(repo):
        branch = repo.as_gitpython().active_branch
        tracking_branch = branch.tracking_branch()
        remote_info = None
        for remote in repo.as_gitpython().remotes:
            pass
            if remote.name == tracking_branch.remote_name:
                remote_info = repo._remote_info(remote)
                break
        return remote_info

    @property
    def active_tracking_remote_head(repo):
        branch = repo.as_gitpython().active_branch
        tracking_branch = branch.tracking_branch()
        return tracking_branch.remote_head

    @property
    def active_tracking_branch_name(repo):
        branch = repo.as_gitpython().active_branch
        tracking_branch = branch.tracking_branch()
        return tracking_branch.name

    @property
    def branches(repo):
        gitrepo = repo.as_gitpython()
        return [branch.name for branch in gitrepo.branches]

    # --- </GIT PYTHON STUFF> ---

    @property
    def aliases(repo):
        aliases = []
        if repo._modname is not None:
            aliases.append(repo._modname)
        aliases.extend(repo._modname_hints[:])
        # if repo.dpath and exists(repo.dpath):
        #     reponame = repo._find_modname_from_repo()
        #     if reponame is not None:
        #         aliases.append(reponame)
        aliases.append(repo.reponame)
        aliases.append(repo.reponame.lower())
        import utool as ut
        aliases = ut.unique(aliases)
        return aliases

    def __nice__(repo):
        reponame = repo.reponame
        modname = repo.modname
        # if modname is False:
        #     print(repo.__dict__)
        if modname is None or modname == reponame:
            return '(%s)' % (reponame,)
        else:
            return '(%s, %s)' % (reponame, modname)

    @property
    def modname(repo):
        # import utool as ut
        modname = None
        if repo._modname is not None:
            modname = repo._modname
        elif len(repo._modname_hints) == 1:
            modname = repo._modname_hints[0]
        else:
            modname = repo.aliases[0]
        return modname

    @property
    def reponame(repo):
        if repo.dpath is not None:
            reponame = basename(repo.dpath)
        elif repo.url is not None:
            url_parts = re.split('[/:]', repo.url)
            reponame = url_parts[-1].replace('.git', '')
        elif repo._modname_hints:
            reponame = repo._modname_hints[0]
        else:
            raise Exception('No way to infer (or even guess) repository name!')
        return reponame

    def _find_modname_from_repo(repo):
        import utool as ut
        packages = ut.get_submodules_from_dpath(repo.dpath, only_packages=True,
                                                recursive=False)
        if len(packages) == 1:
            modname = ut.get_modname_from_modpath(packages[0])
            return modname

    def add_script(repo, key, script):
        repo.scripts[key] = script

    def clone(repo, recursive=False):
        print('[git] check repo exists at %s' % (repo.dpath))
        if recursive:
            args = '--recursive'
        else:
            args = ''
        if not exists(repo.dpath):
            _cd(dirname(repo.dpath))
            print('repo.default_branch = %r' % (repo.default_branch,))
            if repo.default_branch is not None:
                args += ' -b {}'.format(repo.default_branch)
            _syscmd('git clone {args} {url}'.format(args=args, url=repo.url))

    def owner(repo):
        url_parts = re.split('[/:]', repo.url)
        owner = url_parts[-2]
        return owner

    def is_owner(repo, userid):
        return userid is not None and repo.owner.lower() == userid.lower()

    def _fix_url(repo, userid, permitted_repos=[]):
        is_owner = repo.is_owner(userid)
        is_contrib = repo.reponame in permitted_repos
        if is_owner or is_contrib:
            repo._ensure_ssh_format()

    def change_url_format(repo, out_type='ssh'):
        """ Changes the url format for committing """
        url = repo.url
        url_parts = re.split('[/:]', url)
        in_type = url_parts[0]
        url_fmts = {
            'https': ('.com/', 'https://'),
            'ssh':   ('.com:', 'git@'),
        }
        url_fmts['git'] = url_fmts['ssh']
        new_repo_url = url
        for old, new in zip(url_fmts[in_type], url_fmts[out_type]):
            new_repo_url = new_repo_url.replace(old, new)
        # Inplace change
        repo.url = new_repo_url
        print('new format repo.url = {!r}'.format(repo.url))

    def check_importable(repo):
        import utool as ut
        # import utool as ut
        found = False
        tried = []
        errors = []
        for modname in repo.aliases:
            tried.append(modname)
            try:
                ut.import_modname(modname)
            except ImportError as ex:  # NOQA
                tried[-1] += ' but got ImportError'
                errors.append(ex)
                pass
            except AttributeError as ex:  # NOQA
                tried[-1] += ' but got AttributeError'
                errors.append(ex)
            else:
                found = True
                errors.append(None)
                tried[-1] += ' and it worked'
                break
        msg = 'tried %s' % (', '.join(tried))
        return found, msg, errors

    def is_cloned(repo):
        from os.path import exists
        if not exists(repo.dpath):
            return False

    def check_installed(repo):
        import utool as ut
        found = None
        tried = []
        for modname in repo.aliases:
            tried.append(modname)
            found = ut.check_module_installed(modname)
            if found:
                break
        msg = 'tried %s' % (', '.join(tried))
        return found, msg

    def check_cpp_build(repo):
        import utool as ut
        script = repo.get_script('build')
        if script.is_fpath_valid():
            if repo.modname == 'pyflann':
                return True, 'cant detect flann cpp'
            # hack, this doesnt quite do it
            pat = '*' + ut.util_cplat.get_pylib_ext()
            dynlibs = ut.glob(repo.dpath + '/' + repo.modname, pat, recursive=True)
            msg = 'Could not find any dynamic libraries'
            flag = len(dynlibs) > 0
        else:
            flag = True
            msg = 'passed, but didnt expect anything'
        return flag, msg

    def get_script(repo, type_):
        class Script(object):
            def __init__(script):
                script.type_ = type_
                script.text = None
                script.fpath = None
                script.cmake = None

            def is_fpath_valid(script):
                return script.fpath is not None and exists(script.fpath)

            def is_valid(script):
                return script.text or script.is_fpath_valid()

            def exec_(script):
                import utool as ut
                print('+**** exec %s script *******' % (script.type_))
                print('repo = %r' % (repo,))
                with ut.ChdirContext(repo.dpath):
                    if script.is_fpath_valid():
                        normbuild_flag = '--no-rmbuild'
                        if ut.get_argflag(normbuild_flag):
                            ut.cmd(script.fpath + ' ' + normbuild_flag)
                        else:
                            ut.cmd(script.fpath)
                    else:
                        if script.text is not None:
                            print('ABOUT TO EXECUTE')
                            ut.print_code(script.text, 'bash')
                            if ut.are_you_sure('execute above script?'):
                                from os.path import join
                                scriptdir = ut.ensure_app_resource_dir('utool',
                                                                       'build_scripts')
                                script_path = join(scriptdir,
                                                   'script_' + script.type_ + '_' +
                                                   ut.hashstr27(script.text) + '.sh')
                                ut.writeto(script_path, script.text)
                                _ = ut.cmd('bash ', script_path)  # NOQA
                        else:
                            print("CANT QUITE EXECUTE THIS YET")
                            ut.print_code(script.text, 'bash')
                #os.system(scriptname)
                print('L**** exec %s script *******' % (script.type_))

        script = Script()
        script.text = repo.scripts.get(type_, None)

        if script.text is None and type_ == 'build' and repo.dpath:
            if sys.platform.startswith('win32'):
                # vtool --rebuild-sver didnt work with this line
                #scriptname = './mingw_build.bat'
                fpath = join(repo.dpath, 'mingw_build.bat')
            else:
                fpath = join(repo.dpath, 'unix_build.sh')
            if exists(fpath):
                script.fpath = fpath

            cmake = join(repo.dpath, 'CMakeLists.txt')
            if exists(cmake):
                script.cmake = cmake

        return script

    def has_script(repo, type_):
        return repo.get_script(type_).is_valid()

    def custom_build(repo):
        script = repo.get_script('build')
        if script is not None:
            script.exec_()

    def custom_install(repo):
        script = repo.get_script('install')
        if script is not None:
            script.exec_()
        # TODO:
        # import utool as ut
        # ut.print_code(repo.install_script, 'bash')

    def issue(repo, command, sudo=False, dry=False, error='raise', return_out=False):
        """
        issues a command on a repo

        CommandLine:
            python -m utool.util_git --exec-repocmd

        Example:
            >>> # DISABLE_DOCTEST
            >>> from utool.util_git import *  # NOQA
            >>> import utool as ut
            >>> repo = dirname(ut.get_modpath(ut, prefer_pkg=True))
            >>> command = 'git status'
            >>> sudo = False
            >>> result = repocmd(repo, command, sudo)
            >>> print(result)
        """
        import utool as ut
        if ut.WIN32:
            assert not sudo, 'cant sudo on windows'
        if command == 'short_status':
            return repo.short_status()
        command_list = ut.ensure_iterable(command)
        cmdstr = '\n        '.join([cmd_ for cmd_ in command_list])
        if not dry:
            print('+--- *** repocmd(%s) *** ' % (cmdstr,))
            print('repo=%s' % ut.color_text(repo.dpath, 'yellow'))
        verbose = True
        with repo.chdir_context():
            ret = None
            for count, cmd in enumerate(command_list):
                if dry:
                    print(cmd)
                    continue
                if not sudo or ut.WIN32:
                    # ret = os.system(cmd)
                    cmdinfo = ut.cmd2(cmd, verbout=True)
                    out, err, ret = ut.take(cmdinfo, ['out', 'err', 'ret'])
                else:
                    # cmdinfo = ut.cmd2('sudo ' + cmd, verbose=1)
                    out, err, ret = ut.cmd(cmd, sudo=True)
                if verbose > 1:
                    print('ret(%d) = %r' % (count, ret,))
                if ret != 0:
                    if error == 'raise':
                        raise Exception('Failed command %r' % (cmd,))
                    elif error == 'return':
                        return out
                    else:
                        raise ValueError('unknown flag error=%r' % (error,))
                if return_out:
                    return out
        if not dry:
            print('L____')

    def chdir_context(repo, verbose=False):
        import utool as ut
        return ut.ChdirContext(repo.dpath, verbose=verbose)

    def pull2(repo, overwrite=True):
        """
        Pulls and automatically overwrites conflict files.
        """
        cmd = 'git pull --no-edit'
        out = repo.issue(cmd, error='return')
        if overwrite and out is not None:
            repo._handle_overwrite_error(out)
            # Retry
            repo.issue(cmd)

    def checkout2(repo, branch, overwrite=True):
        """
        Checkout `branch` and automatically overwrites conflict files.
        """
        cmd = 'git checkout %s' % (branch,)
        out = repo.issue(cmd, error='return')
        if overwrite and out is not None:
            repo._handle_overwrite_error(out)
            repo._handle_abort_merge_rebase(out)
            # Retry
            repo.issue(cmd)

    def _parse_merge_conflict_fpaths(repo, out):
        fpaths = []
        if out is not None:
            for line in out.split('\n'):
                pref = 'CONFLICT (content): Merge conflict in '
                if line.startswith(pref):
                    fpaths.append(join(repo.dpath, line[len(pref):]))
        return fpaths

    def _handle_abort_merge_rebase(repo, out):
        if out.startswith('error: you need to resolve your current index first'):
            try:
                repo.issue('git merge --abort')
            except Exception:
                pass
            try:
                repo.issue('git rebase --abort')
            except Exception:
                pass

    def _handle_overwrite_error(repo, out):
        import utool as ut
        # parse stdout to handle the error
        if out.startswith('error: The following untracked working tree files would be overwritten'):
            print('[ut.git] handling overwrite error')
            lines = out.split('\n')[1:]
            fpaths = []
            for line in lines:
                if line.startswith('Please move or remove them before you can merge'):
                    break
                fpaths.append(join(repo.dpath, line.strip()))
                ut.remove_file_list(fpaths)

    def short_status(repo):
        r"""
        CommandLine:
            python -m utool.util_git short_status

        Example:
            >>> # DISABLE_DOCTEST
            >>> from utool.util_git import *  # NOQA
            >>> import utool as ut
            >>> repo = Repo(dpath=ut.truepath('.'))
            >>> result = repo.short_status()
            >>> print(result)
        """
        import utool as ut
        prefix = repo.dpath
        with ut.ChdirContext(repo.dpath, verbose=False):
            out, err, ret = ut.cmd('git status', verbose=False, quiet=True)
            # parse git status
            is_clean_msg1 = 'Your branch is up-to-date with'
            is_clean_msgs = [
                'nothing to commit, working directory clean',
                'nothing to commit, working tree clean',
            ]
            msg2 = 'nothing added to commit but untracked files present'

            needs_commit_msgs = [
                'Changes to be committed',
                'Changes not staged for commit',
                'Your branch is ahead of',
            ]

            suffix = ''
            if is_clean_msg1 in out and any(msg in out for msg in is_clean_msgs):
                suffix += ut.color_text('is clean', 'blue')
            if msg2 in out:
                suffix += ut.color_text('has untracked files', 'yellow')
            if any(msg in out for msg in needs_commit_msgs):
                suffix += ut.color_text('has changes', 'red')
        print(prefix + ' ' + suffix)

    def python_develop(repo):
        import utool as ut
        repo.issue('{pythoncmd} -m pip install -e {dpath}'.format(
            pythoncmd=repo.pythoncmd, dpath=repo.dpath),
            sudo=not ut.in_virtual_env())

    def is_gitrepo(repo):
        gitdir = join(repo.dpath, '.git')
        return exists(gitdir) and isdir(gitdir)

    def pull(repo, has_submods=False):
        print('Pulling: ' + repo.dpath)
        _cd(repo.dpath)
        assert repo.is_gitrepo(), 'cannot pull a nongit repo'
        _syscmd('git pull')
        if has_submods:
            _syscmd('git submodule init')
            _syscmd('git submodule update')

    def rename_branch(repo, old_branch_name, new_branch_name, remote='origin'):
        r"""
        References:
            http://stackoverflow.com/questions/1526794/rename?answertab=votes#tab-top
            http://stackoverflow.com/questions/9524933/renaming-a-branch-in-github

        CommandLine:
            python -m utool.util_git --test-rename_branch --old=mymaster --new=ibeis_master

        Example:
            >>> # DISABLE_DOCTEST
            >>> # SCRIPT
            >>> from utool.util_git import *  # NOQA
            >>> repo = ut.get_argval('--repo', str, '.')
            >>> remote = ut.get_argval('--remote', str, 'origin')
            >>> old_branch_name = ut.get_argval('--old', str, None)
            >>> new_branch_name = ut.get_argval('--new', str, None)
            >>> rename_branch(old_branch_name, new_branch_name, repo, remote)
        """
        assert repo.is_gitrepo(), 'cannot pull a nongit repo'
        fmtdict = dict(remote=remote,
                       old_branch_name=old_branch_name,
                       new_branch_name=new_branch_name)
        command_list = [
            'git checkout {old_branch_name}'.format(**fmtdict),
            # rename branch
            'git branch -m {old_branch_name} {new_branch_name}'.format(**fmtdict),
            # delete old branch
            'git push {remote} :{old_branch_name}'.format(**fmtdict),
            # push new branch
            'git push {remote} {new_branch_name}'.format(**fmtdict),
        ]
        repo.issue(command_list)

    @staticmethod
    def resolve_conflicts(fpath, strat, force=False, verbose=True):
        """
        Parses merge conflits and takes either version
        """
        import utool as ut
        import re
        top_pat = re.escape('<' * 7)
        mid_pat = re.escape('=' * 7)
        bot_pat = re.escape('>' * 7)
        flags = re.MULTILINE | re.DOTALL
        # Pattern to remove the top part
        theirs_pat1 = re.compile('^%s.*?%s.*?$\n' % (top_pat, mid_pat), flags=flags)
        theirs_pat2 = re.compile('^%s.*?$\n' % (bot_pat), flags=flags)
        # Pattern to remove the bottom part
        ours_pat1   = re.compile('^%s.*?%s.*?$\n' % (mid_pat, bot_pat), flags=flags)
        ours_pat2   = re.compile('^%s.*?$\n' % (top_pat), flags=flags)
        strat_pats = {
            'theirs': [theirs_pat1, theirs_pat2],
            'ours': [ours_pat1, ours_pat2],
        }

        text_in = ut.readfrom(fpath)
        text_out = text_in
        strat = 'ours'
        strat = 'theirs'
        for pat in strat_pats[strat]:
            text_out = pat.sub('', text_out)
        if verbose:
            ut.print_difftext(ut.difftext(text_in, text_out, num_context_lines=3))

        if force:
            ut.writeto(fpath, text_out)


def git_sequence_editor_squash(fpath):
    r"""
    squashes wip messages

    CommandLine:
        python -m utool.util_git --exec-git_sequence_editor_squash

    Example:
        >>> # DISABLE_DOCTEST
        >>> # SCRIPT
        >>> import utool as ut
        >>> from utool.util_git import *  # NOQA
        >>> fpath = ut.get_argval('--fpath', str, default=None)
        >>> git_sequence_editor_squash(fpath)

    Ignore:
        text = ut.codeblock(
            '''
            pick 852aa05 better doctest for tips
            pick 3c779b8 wip
            pick 02bc21d wip
            pick 1853828 Fixed root tablename
            pick 9d50233 doctest updates
            pick 66230a5 wip
            pick c612e98 wip
            pick b298598 Fixed tablename error
            pick 1120a87 wip
            pick f6c4838 wip
            pick 7f92575 wip
            ''')

    Ignore:
        def squash_consecutive_commits_with_same_message():
            # http://stackoverflow.com/questions/8226278/git-alias-to-squash-all-commits-with-a-particular-commit-message
            # Can do interactively with this. Can it be done automatically and pay attention to
            # Timestamps etc?
            git rebase --interactive HEAD~40 --autosquash
            git rebase --interactive $(git merge-base HEAD master) --autosquash

            # Lookbehind correct version
            %s/\([a-z]* [a-z0-9]* wip\n\)\@<=pick \([a-z0-9]*\) wip/squash \2 wip/gc

           # THE FULL NON-INTERACTIVE AUTOSQUASH SCRIPT
           # TODO: Dont squash if there is a one hour timedelta between commits

           GIT_EDITOR="cat $1" GIT_SEQUENCE_EDITOR="python -m utool.util_git --exec-git_sequence_editor_squash \
                   --fpath $1" git rebase -i $(git rev-list HEAD | tail -n 1) --autosquash --no-verify
           GIT_EDITOR="cat $1" GIT_SEQUENCE_EDITOR="python -m utool.util_git --exec-git_sequence_editor_squash \
                   --fpath $1" git rebase -i HEAD~10 --autosquash --no-verify

           GIT_EDITOR="cat $1" GIT_SEQUENCE_EDITOR="python -m utool.util_git --exec-git_sequence_editor_squash \
                   --fpath $1" git rebase -i $(git merge-base HEAD master) --autosquash --no-verify

           # 14d778fa30a93f85c61f34d09eddb6d2cafd11e2
           # c509a95d4468ebb61097bd9f4d302367424772a3
           # b0ffc26011e33378ee30730c5e0ef1994bfe1a90
           # GIT_SEQUENCE_EDITOR=<script> git rebase -i <params>
           # GIT_SEQUENCE_EDITOR="echo 'FOOBAR $1' " git rebase -i HEAD~40 --autosquash
           # git checkout master
           # git branch -D tmp
           # git checkout -b tmp
           # option to get the tail commit
           $(git rev-list HEAD | tail -n 1)
           # GIT_SEQUENCE_EDITOR="python -m utool.util_git --exec-git_sequence_editor_squash \
                   --fpath $1" git rebase -i HEAD~40 --autosquash
           # GIT_SEQUENCE_EDITOR="python -m utool.util_git --exec-git_sequence_editor_squash \
                   --fpath $1" git rebase -i HEAD~40 --autosquash --no-verify
           <params>
    """
    # print(sys.argv)
    import utool as ut
    text = ut.read_from(fpath)
    # print('fpath = %r' % (fpath,))
    print(text)
    # Doesnt work because of fixed witdth requirement
    # search = (ut.util_regex.positive_lookbehind('[a-z]* [a-z0-9]* wip\n') + 'pick ' +
    #           ut.reponamed_field('hash', '[a-z0-9]*') + ' wip')
    # repl = ('squash ' + ut.bref_field('hash') + ' wip')
    # import re
    # new_text = re.sub(search, repl, text, flags=re.MULTILINE)
    # print(new_text)
    prev_msg = None
    prev_dt = None
    new_lines = []

    def get_commit_date(hashid):
        out, err, ret = ut.cmd('git show -s --format=%ci ' + hashid, verbose=False, quiet=True, pad_stdout=False)
        # from datetime import datetime
        from dateutil import parser
        # print('out = %r' % (out,))
        stamp = out.strip('\n')
        # print('stamp = %r' % (stamp,))
        dt = parser.parse(stamp)
        # dt = datetime.strptime(stamp, '%Y-%m-%d %H:%M:%S %Z')
        # print('dt = %r' % (dt,))
        return dt

    for line in text.split('\n'):
        commit_line = line.split(' ')
        if len(commit_line) < 3:
            prev_msg = None
            prev_dt = None
            new_lines += [line]
            continue
        action = commit_line[0]
        hashid = commit_line[1]
        msg = ' ' .join(commit_line[2:])
        try:
            dt = get_commit_date(hashid)
        except ValueError:
            prev_msg = None
            prev_dt = None
            new_lines += [line]
            continue
        orig_msg = msg
        can_squash = action == 'pick' and msg == 'wip' and prev_msg == 'wip'
        if prev_dt is not None and prev_msg == 'wip':
            tdelta = dt - prev_dt
            # Only squash closely consecutive commits
            threshold_minutes = 45
            td_min = (tdelta.total_seconds() / 60.)
            # print(tdelta)
            can_squash &= td_min < threshold_minutes
            msg = msg + ' -- tdelta=%r' % (ut.get_timedelta_str(tdelta),)
        if can_squash:
            new_line = ' ' .join(['squash', hashid, msg])
            new_lines += [new_line]
        else:
            new_lines += [line]
        prev_msg = orig_msg
        prev_dt = dt
    new_text = '\n'.join(new_lines)

    def get_commit_date(hashid):
        out = ut.cmd('git show -s --format=%ci ' + hashid, verbose=False)
        print('out = %r' % (out,))

    # print('Dry run')
    # ut.dump_autogen_code(fpath, new_text)
    print(new_text)
    ut.write_to(fpath, new_text, n=None)


def std_build_command(repo='.'):
    """
    DEPRICATE
    My standard build script names.

    Calls mingw_build.bat on windows and unix_build.sh  on unix
    """
    import utool as ut
    print('+**** stdbuild *******')
    print('repo = %r' % (repo,))
    if sys.platform.startswith('win32'):
        # vtool --rebuild-sver didnt work with this line
        #scriptname = './mingw_build.bat'
        scriptname = 'mingw_build.bat'
    else:
        scriptname = './unix_build.sh'
    if repo == '':
        # default to cwd
        repo = '.'
    else:
        os.chdir(repo)
    ut.assert_exists(scriptname)
    normbuild_flag = '--no-rmbuild'
    if ut.get_argflag(normbuild_flag):
        scriptname += ' ' + normbuild_flag
    # Execute build
    ut.cmd(scriptname)
    #os.system(scriptname)
    print('L**** stdbuild *******')


if __name__ == '__main__':
    r"""
    CommandLine:
        python -m utool.util_git
        python -m utool.util_git --allexamples
    """
    import multiprocessing
    multiprocessing.freeze_support()  # for win32
    import utool as ut  # NOQA
    ut.doctest_funcs()
